#!/usr/local/bin/perl -w

use strict;

use Carp;

use Vector;


=head1

    The goal is to be able to determine if a quad is facing the viewer
or is a backface for the viewer so that OpenGL backface culling and
shading work correctly.

    The method for testing this is to shoot a ray from infinity (to
avoid perspective problems) at a point in the quad and count the
number of quads the ray passes through.  If the number is odd then the
quad is front facing, if even the quad is back facing.

    The odd case is when a quad is parallel to the ray.  If the
parallel quad is not the target quad then the count should not be
affected, if target quad is parallel to the ray then there is a
problem and this is a "side" of the object and I am not sure how to
work out the facing.

    The complexity will grow by O(N^2) where N is the number of quads.

    Another way to do this is a flow method comparing the direction of
the quad ordering to that of its neigboring quad.  So we could use the
ray test to get the correct facing of the first quad and then make all
of the neighbors correct with respect to this quad, then recurse
through the remaining quads until all are checked.  The catch is
determining neighbors and what to do if the quads are not all
connected.  The fix to the latter is to flow through until all
connected quads are tested, then see if all quads are tested.  If
there are remaining quads then shoot a ray to one to get the correct
facing for that quad thenflow through all connected quads until you
are out of unchecked quads.

    The complexity for this is O(N * P) where N is the number of quads and
P is the number of connected parts (so if all quads are connected P = 1).

=cut

# Use the flow method to check the objects
sub correct_poly_facing {
    my $objs = shift;
    
    # Set up a datastructure to hold the line segments defined by the
    #  2 points at each end.
    # This will be a hash to hashrefs of hashrefs of hashrefs ...
    #  to an array ref listing the object refs
    # e.g. for points 1 & 2:
    #  $segs{$1X}{$1Y}{$1Z}{$2X}{$2Y}{$2Z} = [];
    my $segs = $objs->get_segs;

    # I am going to arbitrarily decide to start with the first quad
    # in the part and assume that it is outward facing and set up 
    # correctly
    # TODO: Make sure it is CCW and correct it if it is not
    my $start = $objs->get_objs_by_type->[4][0];

    # Now walk all of the polygons and see if they are facing the same 
    # way as their neighbors.  Keep a hash of all of the object references
    # so that we know if we have visited the objects.  Note that this
    # will recurse through all connected polys
    my $visited = {};
    while (defined $start) {
	print "NEW START\n";

	check_neighbors($start, $segs, $visited);
	
	# TODO: This should check to see if the new $start is facing 
	# the right way

	$start = (grep { not exists $visited->{$_} }
		  @{$objs->get_objs_by_type->[4]})[0];
    }
    
}

# Check facing of quad against the neighbors
sub check_neighbors {
    my $obj = shift or die "Bad obj";
    my $segs = shift or die "Bad segs";
    my $visited = shift or die "Bad visited";

    
    return 
	if defined $visited->{$obj} and $visited->{$obj} == 1;

    $visited->{$obj} = 1;


    my $N = $obj->points;
    my $n = int @$N;
    my $i;
    for ($i = 0; $i < $n; $i++) {
	my $j = ($i + 1) % $n;

	# Get the two points
	my $p1 = $N->[$i];
	my $p2 = $N->[$j];

	# Count the number of polys attached to the line segment
	my $c1 = 0;
	$c1 = int( @{$segs->{$p1->x}{$p1->y}{$p1->z}
		     ->{$p2->x}{$p2->y}{$p2->z}} )
	    if defined( @{$segs->{$p1->x}{$p1->y}{$p1->z}
			  ->{$p2->x}{$p2->y}{$p2->z}} );
	
	my $c2 = 0;
	$c1 = int( @{$segs->{$p2->x}{$p2->y}{$p2->z}
		     ->{$p1->x}{$p1->y}{$p1->z}} )
	    if defined( @{$segs->{$p2->x}{$p2->y}{$p2->z}
			  ->{$p1->x}{$p1->y}{$p1->z}} );

	# See if there are too many polys defined by the line segment
	if ($c1 + $c2 != 2) {
	    print STDERR "Bad count $c1+$c2: " .
		join(":", $N->[$i]->data_asrow), " to " .
		join(":", $N->[$j]->data_asrow), " for poly " .
		$obj->print, "\n";
	    next;
	}
	
	# Check that the two polys go in the same direction
	if ($c1 > 1) {
	    my $poly;
	    foreach $poly (@{$segs->{$p1->x}{$p1->y}{$p1->z}
			          ->{$p2->x}{$p2->y}{$p2->z}})
	    {
		# Ignore ourselves
		next if $poly == $obj;

		# Correct the neighbor by reordering the points
		# Update the segment list to reflect the change
		correct_neighbor($poly, $segs, $visited);
	    }
	}
	
	# Now check all of the neighbors
	my $neighbor;
	foreach $neighbor 
	    (@{$segs->{$p1->x}{$p1->y}{$p1->z}->{$p2->x}{$p2->y}{$p2->z}},
	     @{$segs->{$p2->x}{$p2->y}{$p2->z}->{$p1->x}{$p1->y}{$p1->z}})
	    {
		check_neighbors($neighbor, $segs, $visited)
		    if defined $obj and 
		       defined $neighbor and 
		       $neighbor != $obj;
	    }
    }
    
}

sub correct_neighbor {
    my $poly = shift;
    my $segs = shift;
    my $visited = shift;

    # Make sure the neighbor has not been visited
    if ($visited->{$poly}) {
	print STDERR "Error: poly ", $poly->print . 
	    " is involved in a twist\n";
	return;
    }

    print STDERR "Fixing: ", $poly->print, "\n";
    
    # Change the ordering of the neighbor points
    my $N = $poly->points;
    my $n = int @$N;

    my ($i, $j);
    for ($i = 0, $j = $n-1; $i < $j; $i++, $j--) {
	# Swap element $i and $j
	my $tmp = $poly->points->[$i];
	$poly->points->[$i] = $poly->points->[$j];
	$poly->points->[$j] = $tmp;
    }
    
    # Fix the neighbor normal
    $poly->set_normal($poly->normal->reverse);
    
    # Now update the segment list to reflect the correction
    for ($i = 0; $i < $n; $i++) {
	my $j = ($i + 1) % $n;
	
	# Get the two points
	my $p1 = $N->[$i];
	my $p2 = $N->[$j];

	# Remove the object from the $p2, $p1 segment
	$segs->{$p2->x}{$p2->y}{$p2->z}->{$p1->x}{$p1->y}{$p1->z} = 
	    [ grep { $_ != $poly }
	      @{$segs->{$p2->x}{$p2->y}{$p2->z}->{$p1->x}{$p1->y}{$p1->z}} ];

	# Add the object to the $p1, $p2 segment
	push(@{$segs->{$p1->x}{$p1->y}{$p1->z}->{$p2->x}{$p2->y}{$p2->z}},
	     $poly);
    }

}


{
    my $objs;

    sub ray_poly_facing {
	my $dl = shift;

	$objs = $dl;
	
	# Draw the object by walking it with drawing actions
	my $actions = {1 => undef,
		       3 => \&check_poly_facing,
		       4 => \&check_poly_facing,
		   };
	
	$dl->walk($actions);
    }


# This will check the given object against all other objects in the part
# Should be called from walk
# This implements the ray method
sub check_poly_facing {
    my $obj = shift;

    my ($x, $y, $z) = (0, 1, 2);

    # Is the object currently a front?
    # Look at the normal and see if the Z component is > 0, 
    # if so it is a front, if < 0 it is a back, and if == 0 it is a problem
    my $normal = $obj->normal->get($z);
    
    if ($normal == 0) { 
	# Handle the perpendicular case by changing to look at the y axis
	$normal = $obj->normal->get($y);
	($x, $y, $z) = (0, 2, 1);
    }
    
    # Get the correct facing
    my $pfp = isFrontObject($obj, $objs, $x, $y, $z);

    if ($normal > 0 and $pfp == -1 or
	$normal < 0 and $pfp == 1) 
    {
	print STDERR "Pre : ", $obj->print, "\n";
	# Change the ordering of the points
	my $N = int @{$obj->points};
	my ($i, $j);
	for ($i = 0, $j = $N-1; $i < $j; $i++, $j--) {
	    
	    # Swap element $i and $j
	    my $tmp = $obj->points->[$i];
	    $obj->points->[$i] = $obj->points->[$j];
	    $obj->points->[$j] = $tmp;
	}
	print STDERR "Post: ", $obj->print, "\n";
	
	# Fix the normal
	$obj->set_normal($obj->normal->reverse);

	# We are flipped so fix it
	print STDERR "Fix it! ";
    }
    else {
	print STDERR "Fine! ";
    }
    print STDERR "$normal $pfp: ", $obj->print, "\n";
}
}

# Purpose: Get a point that definitely exists in the polygon
#
# Arg: polygon (LDObj of type 3 or 4)
#
# Return: Vector of size 3 with the X, Y amd Z coords of a point 
#   in the polygon
sub findPointInPolygon {
    my $poly = shift;

    croak "Can not find a point in anything other than a triangle " .
	"or quad " . $poly->type . ": " . $poly->raw_line
	    unless $poly->type == 3 or $poly->type == 4;
    
    # Take the average of the XYZ components of the points
    my @count = (0, 0, 0);
    my $p;
    foreach $p (@{$poly->points}) {
	my $i;
	foreach $i (0..2) {
	    $count[$i] += $p->get($i);
	}
    }
    my $i;
    foreach $i (0..2) {
	$count[$i] /= $poly->type;
    }

    return Vector->new("-n" => 3, "-data" => \@count);
}

# Purpose: Does point exist in polygon?
#          The optional x, y, and z arguments allow you to change
#          the orientation so we can check along different axes
# In other words if I shoot a ray from (point.x, point.y, +infinity) to
# the point, will it pass through the polygon?
# 
# Args: point (Vector)
#       polygon (LDObj of type 3 or 4)
#       x (int optional, default 0)
#       y (int optional, default 1)
#       z (int optional, default 2)
#
# Return: 1 for true
#         0 for false
#         -1 for perpendicular
sub isPointBehindPolygon {
    my $point = shift;
    my $poly = shift;

    my $x = shift; $x = 0 unless defined $x;
    my $y = shift; $y = 1 unless defined $y;
    my $z = shift; $z = 2 unless defined $z;

    croak "Can not test for a point in anything other than a triangle " .
	"or quad " . $poly->type . ": " . $poly->raw_line
	    unless $poly->type == 3 or $poly->type == 4;

    # First see if the polygon is parallel to the Z axis
    # We do this by checking that the normal to the polygon
    # is not in the X, Y plane
    return -1 
	if $poly->normal->get($z) == 0;


    # Is the point behind the polygon?
    
    # Find the largest & smallest Z components of the polygon
    my $maxZ = undef;
    my $minZ = undef;
    my $p;
    foreach $p (@{$poly->points}) {
	$maxZ = max($p->get($z), $maxZ);
	$minZ = min($p->get($z), $minZ);
    }

    # Now check that the max Z component of the polygon is larger
    # than the Z component of the point we are checking against
    #print STDERR "MZ: $maxZ   PZ: ", $point->get($z), "\n";
    return 0 
	unless $maxZ > $point->get($z);

    # Now we know that we are not in front of the point we can see if 
    # the X & Y of the point is within the X & Y area of the polygon
    return IsPointInPolygon($point, $poly, $x, $y, $z);
}


# Purpose: Check poly against all polygons in the object to see if 
#          is is a front or a back
#
# Args: poly (LDObj) 
#       objects (LDraw)
#
# Returns: 1 is this is a front point
#          0 if it is a back
#          -1 if the question is undefined
sub isFrontObject {
    my $poly = shift;
    my $objects = shift;

    my $x = shift; $x = 0 unless defined $x;
    my $y = shift; $y = 1 unless defined $y;
    my $z = shift; $z = 2 unless defined $z;

    # Get a point in the polygon
    my $point = findPointInPolygon($poly);

    # See what the orientation should be, if the poly is parallel to the
    # Z axis then we use the Y axis instead of the Z
    #
    # We do this by checking that the normal to the polygon
    # is not in the X, Y plane
    if ($poly->normal->get($z) == 0) {
	my $tmp = $z;
	$z = $y;
	$y = $tmp;
    }

    # Get the LDObjs from the LDraw object
    my $types = $objects->get_objs_by_type;

    # Look at each object and see if it is a front or back
    my $front = 0;
    my $other_poly;
    foreach $other_poly (@{$types->[3]}, @{$types->[4]}) {
	my $pfront = isPointBehindPolygon($point, $other_poly, $x, $y, $z);
	#print "PFront: $pfront - ", $poly->print, "\n";

	# If the point is behind the polygon then we need to 
	# increment the count
	$front++
	    if $pfront == 1;
    }

    return 0 # Back
	if ($front % 2 == 0);
    
    return 1; # Front
}


# Returns the largest of the arguments
sub max {
    my $l = shift;
    my $c;
    while (defined ($c = shift)) {
	$l = $c 
	    if $c > $l;
    }

    return $l;
}

# Returns the smallest of the arguments
sub min {
    my $l = shift;
    my $c = 0;
    while (defined ($c = shift)) {
	$l = $c 
	    if $c < $l;
    }

    return $l;
}

# Purpose: Test to see if a point is in a polygon
#
# Args: point (Vector)
#       polygon (LDObj of type 3 or 4)
#       x (int optional, default 0)
#       y (int optional, default 1)
#       z (int optional, default 2)
#
# Return: 1 for true (point is in polygon)
#         0 for false (point is not in polygon)
#
# See http://www.ecse.rpi.edu/Homepages/wrf/geom/pnpoly.html
# See http://www.mhri.edu.au/~pdb/geometry/insidepoly/
sub IsPointInPolygon {
    my $point = shift;
    my $poly = shift;

    my $x = shift; $x = 0 unless defined $x;
    my $y = shift; $y = 1 unless defined $y;
    my $z = shift; $z = 2 unless defined $z;

    # Set up the data
    my $npol = int @{$poly->points};
    my $px = $point->get($x);
    my $py = $point->get($y);

    my @xp = ();
    my @yp = ();
    my ($i, $j, $c) = (0, 0, 0);
    foreach $i (0 .. $npol-1) {
	$xp[$i] = $poly->points->[$i]->get($x);
	$yp[$i] = $poly->points->[$i]->get($y);
    }

    for ($i = 0, $j = $npol-1; $i < $npol; $j = $i++) {
	if (((($yp[$i]<=$y) && ($y<$yp[$j])) ||
	     (($yp[$j]<=$y) && ($y<$yp[$i]))) &&
	    ($x < ($xp[$j] - $xp[$i]) * ($y - $yp[$i]) / 
	     ($yp[$j] - $yp[$i]) + $xp[$i])
	    )
	{
	    $c++;
	}
    }

    if (0) {
	print STDERR "COUNT: $c\n";
	print STDERR "POINT: ", $point->print;
	print STDERR "OBJ: ", $poly->print, "\n";
    }

    return 0 # Outside
	if ($c % 2 == 0);
    
    return 1; # Inside
}

# This is a working point in polygon but I think it is less efficient
sub IsPointInPolygonOther {
    my $point = shift;
    my $poly = shift;

    my $x = shift; $x = 0 unless defined $x;
    my $y = shift; $y = 1 unless defined $y;
    my $z = shift; $z = 2 unless defined $z;

    my $counter = 0;
    my $i;
    my $xinters;
    my ($p1, $p2);

    my @polygon = @{$poly->points};
    my $N = int @polygon;

    $p1 = $polygon[0];
    for ($i = 1; $i <= $N; $i++) {
	$p2 = $polygon[$i % $N];
	if ($point->get($y) > min($p1->get($y), $p2->get($y))) {
	    if ($point->y <= max($p1->y,$p2->y)) {
		if ($point->x <= max($p1->x,$p2->x)) {
		    if ($p1->y != $p2->y) {
			$xinters = ($point->y - $p1->y) *
			    ($p2->x - $p1->x) /
				($p2->y - $p1->y) + $p1->x;

			$counter++
			    if ($p1->x == $p2->x || 
				$point->x <= $xinters)
		    }
		}
	    }
	}
	$p1 = $p2;
    }

    print STDERR "COUNT: $counter\n";
    print STDERR "POINT: ", $point->print;
    print STDERR "OBJ: ", $poly->print, "\n";

    return 0 # Outside
	if ($counter % 2 == 0);
    
    return 1; # Inside

}

1;
