#!/usr/local/bin/perl -w

# This provides a class for representing the displayable objects

package DisplayList;

use strict;
use vars qw(@ISA @EXPORT @EXPORT_OK %EXPORT_TAGS $VERSION);

# Allow carping of errors
use Carp;

use Matrix;

use Vector;

use Exporter;
$VERSION = 1.00;
@ISA = qw(Exporter);

@EXPORT      = qw();
@EXPORT_OK   = qw();
%EXPORT_TAGS = ();


=head1 NAME

DisplayList - Convert an LDraw object into a display list object

=head1 SYNOPSIS

    use DisplayList;

    my $ld = DisplayList->new($ldraw);

=cut

=head2 DisplayList->new($ldraw);

This looks at an LDraw object and converts it into a DisplayList.

This returns a DisplayList object.

=cut

sub new {
    my $class = shift;
    
    # Get the class name if we were called through an object
    $class = ref $class if ref $class;

    # Get the file
    my $ld = shift 
	or croak "no LDraw object given";

    # Set up memory for the object and bless it into the class
    my $self = {};
    bless($self, $class);

    # Initialize the object
    $self->_init($ld);

    # And give the object back to the caller
    return $self;
}


=head2 $displaylist->get_objs_by_type();

This returns the B<raw> list of the types as a arrayref keyed on the type
and the value for each is an arrayref to the LDObj list of all of the
objects of that type.  If you modify this list, do so at your own risk.

=cut
    
sub get_objs_by_type {
    return shift->{"TYPES"};
}


=head2 $displaylist->get_segs();

This returns the B<raw> segment information.

=cut
    
sub get_segs {
    return shift->{"SEGS"};
}


=head2 $ldraw->walk($actions);

This will traverse all of the objects of the specified type and 
run the action for that type.

$actions is a hashref for each type to work on.  Only the types
with a defined action will be walked so:
    $actions = {1 => \&check_obj,
		4 => \&check_quad,
	       };
Will only walk the quads and the objects.  Note that subobjects 
will be checked by default.  If you do not want this behavior 
explicitly send in an undef for type 1, or define your own
object checker instead.

The action subroutine will be passed each LDObj that is 
encountered, and is expected to return 1 if the check is successful 
0 if it fails and to set $! to the reason the check failed. 

This will either return 1 to indicate success.  On failure 0 is
returned and $! will indicate the first error that occured.

=cut

sub walk {
    my $self = shift;
    my $actions = shift;
    croak "instance method called on class" unless ref $self;
    croak "bad actions" unless ref $actions;

    # Walk all the types only if the action is defined
    my $type;
    for ($type = 0; $type < int @{$self->{"TYPES"}}; $type++) {
	if (defined $actions->{$type}) {
	    # Walk the tree
	    my $obj;
	    foreach $obj (@{$self->{"TYPES"}[$type]}) {
		&{$actions->{$type}}($obj, $actions);
	    }
	}
    }

    return 1;
}


####
# PRIVATE functions, do not use unless you absolutely have to
#   I will not guarantee that the below will not change
####

# This will set up the object data structures
# Should be called by subclasses
sub _init {
    my $self = shift;
    my $ld = shift;
    croak "Bad arguments to the constructor"
	unless ref $self and ref $ld;

    # Set up the instance variables
    $self->{"TYPES"} = [];    # Index the LD objects by type number
    $self->{"SEGS"} = {};     # Store the polygon line segments     


    # Walk through the LDraw object and define all of the types
    # that exist in the base file _and the subfiles_ and translate
    # them based on the translation and point data of the file that
    # called them
    my $trans_matrix = Matrix->new(-n => 4, -m => 4, 
				   -data => [[1, 0, 0, 0],
					     [0, 1, 0, 0],
					     [0, 0, 1, 0],
					     [0, 0, 0, 1]]
				   );
    my @translation_stack = ($trans_matrix);
    
    my $color = 16;
    my $populate_callback;
    $populate_callback = sub {
	my $obj = shift;

	# Can be called on types 1, 2, 3, 4
	my $type = $obj->type;

	# Handle the child objects
	if ($type == 1) {
	    my $t = $obj->translation;
	    my $p = $obj->points->[0];
	    my $new_trans = 
		Matrix->new(-m => 4, -n => 4, -data => 
			    [[$t->[0]->data_asrow, $p->get(0)],
			     [$t->[1]->data_asrow, $p->get(1)],
			     [$t->[2]->data_asrow, $p->get(2)],
			     [0,0,0,1]]
			    );

	    # Calculate the new translation matrix based on the last
	    confess "Empty translation stack"
		if int(@translation_stack) == 0;
	    $trans_matrix = $translation_stack[0]->clone;
	    $trans_matrix = $trans_matrix->multiply($new_trans);

	    # Unshift onto the translation stack
	    unshift @translation_stack, $trans_matrix;

	    # Walk the child object
	    $obj->subpart->
		walk({1 => $populate_callback,
		      2 => $populate_callback,
		      3 => $populate_callback,
		      4 => $populate_callback,
		     });

	    # Shift when the walk returns
	    shift @translation_stack
		or confess "Translation stack empty";
	}

	# Translate ourself and store the new object
	elsif ($type >= 2 and  $type <= 4) {
	    my $new = $obj->clone;
	    
	    # Get the current translation
	    $trans_matrix = $translation_stack[0]
		or confess "Translation stack empty";

	    # Translate all of the points and renormalize
	    my $point;
	    foreach $point (@{$new->points}) {
		apply_translation($point, $trans_matrix);
	    }
	    $new->renormalize
		if $type != 2;

	    push @{$self->{"TYPES"}[$type]}, $new;
	}
    };
    # Walk the LDObject
    $ld->walk({1 => $populate_callback,
	       2 => $populate_callback,
	       3 => $populate_callback,
	       4 => $populate_callback,
	      });


    # Walk through all of the objects (in this file) and define the 
    # line segments that are used to compose the vertices.  Note the
    # lines are not looked at since we don't care about them
    my $callback = sub {
	my $obj = shift;
	
	# Load all of the line segments into the hash and push the
	# object ref into the array ref
	my $N = $obj->points;
	my $n = int @$N;
	my $i;
	for ($i = 0; $i < $n; $i++) {
	    my $j = ($i + 1) % $n;
	    
	    push(@{$self->{"SEGS"}->{$N->[$i]->x}{$N->[$i]->y}{$N->[$i]->z}
		                  ->{$N->[$j]->x}{$N->[$j]->y}{$N->[$j]->z}},
		 $obj);
	}
    };

    # Ignore lines when looking for the segments
    $self->walk({1 => undef,
		 3 => $callback,
		 4 => $callback,
	        });
    
    return $self;
}

# Apply the current translation to the given vector
sub apply_translation {
    my $vector = shift;
    my $trans = shift;
    croak "Bad arguments to apply_translation"
	unless ref $vector and ref $trans;

    # The vector is only 3 big so we need to add an element to the end
    my @vec = ($vector->data_asrow, 1);

    # Temporarily store the result
    my @result = ();

    # Multiply the vector (x, y, z, 1) by the translation matrix
    # and return the result (note this is simplified since it is always
    # a matrix * vector)
    my ($a, $b) = (0, 0);
    my $sN = $trans->N - 1;
    my $sM = $trans->M - 1;
    foreach $b (0 .. $sN) {
	$result[$b] = 0;
	foreach $a (0 .. $sM) {
	    $result[$b] += $trans->_get($a, $b) * $vec[$a];
	}
    }

    # Remove the trailing 1 from the result and move the result into the vector
    pop @result;
    $vector->{"DATA"} = \@result;
}

1;
