#!/usr/local/bin/perl -w

# This provides a class for manipulating LDraw file objects

package LDraw;

package LDObj;

use strict;
use vars qw(@ISA @EXPORT @EXPORT_OK %EXPORT_TAGS $VERSION);

# Allow carping of errors
use Carp;

# Use the object oriented filehandle interface
use IO::File;

use Exporter;
$VERSION = 1.00;
@ISA = qw(Exporter);

@EXPORT      = qw();
@EXPORT_OK   = qw();
%EXPORT_TAGS = ();


# Set up the read callbacks
# This is a private classwide variable
my @read_callback = ();
$read_callback[0] = \&read_comment_callback;
$read_callback[1] = \&read_part_callback;
$read_callback[2] = \&read_line_callback;
$read_callback[3] = \&read_triangle_callback;
$read_callback[4] = \&read_quad_callback;
$read_callback[5] = \&read_cond_callback;


sub new {
    my $class = shift;

    # Get the class name if we were called through an object
    $class = ref $class if ref $class;

    # Set up memory for the object and bless it into the class
    my $self = {};
    bless($self, $class);

    # Initialize the object
    $self->_init(@_);

    # And give the object back to the caller
    return $self;
}

sub type {
    my $self = shift;
    croak "instance method called on class" unless ref $self;

    return $self->{"TYPE"};
}

sub subtype {
    my $self = shift;
    croak "instance method called on class" unless ref $self;

    return $self->{"SUBTYPE"};
}

sub parent {
    my $self = shift;
    croak "instance method called on class" unless ref $self;

    return $self->{"PARENT"};
}

sub line_no {
    my $self = shift;
    croak "instance method called on class" unless ref $self;

    return $self->{"LINE_NO"};
}

sub raw_line {
    my $self = shift;
    croak "instance method called on class" unless ref $self;

    return $self->{"RAW_LINE"};
}

sub data {
    my $self = shift;
    croak "instance method called on class" unless ref $self;

    return $self->{"DATA"};
}

sub points {
    my $self = shift;
    croak "instance method called on class" unless ref $self;

    return $self->{"POINT"};
}

sub subpartname {
    my $self = shift;
    croak "instance method called on class" unless ref $self;

    return $self->{"SUBPARTNAME"};
}

sub subpart {
    my $self = shift;
    croak "instance method called on class" unless ref $self;

    return $self->{"SUBPART"};
}

sub color {
    my $self = shift;
    croak "instance method called on class" unless ref $self;

    return $self->{"COLOR"};
}

sub set_color {
    my ($self, $color) = @_;
    croak "instance method called on class" unless ref $self;

    $self->{"COLOR"} = $color;
    return $self;
}


sub translation {
    my $self = shift;
    croak "instance method called on class" unless ref $self;

    return $self->{"TRANSLATION"};
}

sub normal {
    my $self = shift;
    croak "instance method called on class" unless ref $self;

    return $self->{"NORMAL"};
}

sub renormalize {
    my $self = shift;
    croak "instance method called on class" unless ref $self;

    # Calculate the normal
    my $AC = $self->{"POINT"}[1]->subtract($self->{"POINT"}[0]);
    my $AB = $self->{"POINT"}[2]->subtract($self->{"POINT"}[0]);
    $self->{"NORMAL"} = $AC->cross($AB)->normalize;
}

sub set_normal {
    my $self = shift;
    my $normal = shift;
    croak "instance method called on class" unless ref $self;
    croak "no normal specified" unless ref $normal;

    $self->{"NORMAL"} = $normal;

    return $self;
}

sub print {
    my $self = shift;
    croak "instance method called on class" unless ref $self;

    my @a = ();
    my $p;
    foreach $p (@{$self->points}) {
	push @a, join(",",$p->data_asrow),
    }
    
    return "(" . join(") (", @a) . ")";
}

# Make a copy of ourself and return it
sub clone {
    my $self = shift;
    croak "instance method called on class" unless ref $self;

    # Get the class name from the object
    my $class = ref $self;

    # Set up memory for the object and bless it into the class
    my $clone = {};
    bless($clone, $class);

    # Initialize the object
    $clone->_clone($self);

    # And give the object back to the caller
    return $clone;
}

####
# PRIVATE functions, do not use unless you absolutely have to
#   I will not guarantee that the below will not change
####


# Initialize the object
# Should probably be called by subclasses
sub _init {
    my $self = shift;
    my $line = shift;
    my $line_no = shift;
    my $ld = shift;
    croak "Bad arguments to the constructor"
	unless ref $self and 
	    defined $line and 
	    defined $line_no and
	    ref $ld;

    # Set up the instance variables
    $self->{"TYPE"} = undef;
    $self->{"RAW_LINE"} = $line;
    $self->{"LINE_NO"} = $line_no;
    $self->{"SUBTYPE"} = undef;
    $self->{"DATA"} = undef;
    $self->{"POINT"} = [];
    $self->{"SUBPARTNAME"} = undef;
    $self->{"SUBPART"} = undef;
    $self->{"COLOR"} = undef;
    $self->{"TRANSLATION"} = undef;
    $self->{"NORMAL"} = undef;
    $self->{"PARENT"} = $ld;
    
    # Parse the input line
    $self->parse($line);
    
    return $self;
}

# Clone the object
# Should probably be called by subclasses
sub _clone {
    my $self = shift;
    my $other = shift;
    croak "Bad arguments to the constructor"
	unless ref $self and ref $other;

    # Copy the instance variables
    my $key;
    foreach $key (keys %$other) {
	$self->{$key} = $other->{$key};
    }
    $self->{"NORMAL"} = $other->{"NORMAL"}->clone
	if defined $other->{"NORMAL"};

    # Copy the arrays of vectors
    foreach $key ("TRANSLATION", "POINT") {
	if (defined $other->{$key}) {
	    $self->{$key} = [];
	    my $i;
	    for ($i = 0; $i < int @{$other->{$key}}; $i++) {
		$self->{$key}->[$i] = $other->{$key}->[$i]->clone;
	    }
	}
    }

    return $self;
}

# Parse an individual LDraw line
sub parse {
    my ($self, $line) = @_;

    # See if the line is blank
    if ($line =~ m|^\s*$|) {
	$self->{"TYPE"} = -1;
	$self->{"SUBTYPE"} = "BLANK";
	return 1;
    }

    # Parse the line
    my ($type, $rest);
    if ($line =~ m|^\s*(\d)\s+(.*)$|) {
	$type = $1;
	$rest = $2;
    }

    # Check the type and callback list
    if (not defined $type or 
	not defined $read_callback[$type]) 
    {
	croak "Bad $type line: $line";
    }
    
    # Call the callback
    $self->{"TYPE"} = $type;
    return &{$read_callback[$type]}($self, $rest);
}


sub read_comment_callback {
    my ($self, $rest) = @_;

    my $t = "COMM";

    if ($self->line_no == 0) {
	$t = "TITLE";
    }
    else {
	my $first = "";
	$first = uc $1 if ($rest =~ m|^\s*(\S+)|);
	
	$t = "\u$first"
	    if $first =~ m\^STEP|WRITE|PRINT|CLEAR|PAUSE|SAVE$\;
    }

    $self->{"SUBTYPE"} = $t;
    $self->{"DATA"} = $rest;

    return 1;
}

sub read_part_callback {
    my ($self, $rest) = @_;
    
    my %v = parse_line("n     f f f f  f  f  f  f  f  f  f  f  s",
		       "color x y z Xx Xy Xz Yx Yy Yz Zx Zy Zz part", 
		       $rest);

    $self->{"SUBTYPE"} = "SUBPART";
    $self->{"POINT"}[0] = Vector->new("-data" => [$v{x}, $v{y}, $v{z}]);
    $self->{"COLOR"} = $v{color};

    $self->{"TRANSLATION"}[0] = 
	Vector->new("-data" => [$v{Xx}, $v{Xy}, $v{Xz}]);
    $self->{"TRANSLATION"}[1] = 
	Vector->new("-data" => [$v{Yx}, $v{Yy}, $v{Yz}]);
    $self->{"TRANSLATION"}[2] = 
	Vector->new("-data" => [$v{Zx}, $v{Zy}, $v{Zz}]);

    $self->{"SUBPARTNAME"} = $v{part};
    $self->{"SUBPART"} = LDraw->new($v{part});

    return 1;
}

sub read_line_callback {
    my ($self, $rest) = @_;
    
    my %v = parse_line("n     f  f  f  f  f  f",
		       "color x1 y1 z1 x2 y2 z2", 
		       $rest);

    $self->{"SUBTYPE"} = "LINE";
    $self->{"COLOR"} = $v{color};
    $self->{"POINT"}[0] = Vector->new("-data" => [$v{x1}, $v{y1}, $v{z1}]);
    $self->{"POINT"}[1] = Vector->new("-data" => [$v{x2}, $v{y2}, $v{z2}]);

    return 1;
}

sub read_triangle_callback {
    my ($self, $rest) = @_;
    
    my %v = parse_line("n     f  f  f  f  f  f  f  f  f",
		       "color x1 y1 z1 x2 y2 z2 x3 y3 z3", 
		       $rest);

    $self->{"SUBTYPE"} = "TRI";
    $self->{"COLOR"} = $v{color};
    $self->{"POINT"}[0] = Vector->new("-data" => [$v{x1}, $v{y1}, $v{z1}]);
    $self->{"POINT"}[1] = Vector->new("-data" => [$v{x2}, $v{y2}, $v{z2}]);
    $self->{"POINT"}[2] = Vector->new("-data" => [$v{x3}, $v{y3}, $v{z3}]);

    # Calculate the normal
    $self->renormalize;

    return 1;
}

sub read_quad_callback {
    my ($self, $rest) = @_;
    
    my %v = parse_line("n     f  f  f  f  f  f  f  f  f  f  f  f",
		       "color x1 y1 z1 x2 y2 z2 x3 y3 z3 x4 y4 z4", 
		       $rest);

    # Set up the points
    $self->{"SUBTYPE"} = "QUAD";
    $self->{"COLOR"} = $v{color};
    $self->{"POINT"}[0] = Vector->new("-data" => [$v{x1}, $v{y1}, $v{z1}]);
    $self->{"POINT"}[1] = Vector->new("-data" => [$v{x2}, $v{y2}, $v{z2}]);
    $self->{"POINT"}[2] = Vector->new("-data" => [$v{x3}, $v{y3}, $v{z3}]);
    $self->{"POINT"}[3] = Vector->new("-data" => [$v{x4}, $v{y4}, $v{z4}]);

    # Make sure the quad is valid, this will calculate the normal as a freebie
    $self->check_quad;
    
    return 1;
}

sub read_cond_callback {
    my ($self, $rest) = @_;
    
    my %v = parse_line("n     f  f  f  f  f  f  f  f  f  f  f  f",
		       "color x1 y1 z1 x2 y2 z2 x3 y3 z3 x4 y4 z4", 
		       $rest);

    # Set up the points
    $self->{"SUBTYPE"} = "COND";
    $self->{"COLOR"} = $v{color};
    $self->{"POINT"}[0] = Vector->new("-data" => [$v{x1}, $v{y1}, $v{z1}]);
    $self->{"POINT"}[1] = Vector->new("-data" => [$v{x2}, $v{y2}, $v{z2}]);
    $self->{"POINT"}[2] = Vector->new("-data" => [$v{x3}, $v{y3}, $v{z3}]);
    $self->{"POINT"}[3] = Vector->new("-data" => [$v{x4}, $v{y4}, $v{z4}]);

    return 1;
}

# Take a format line and break up a data line and add it to a hash with 
# the names from the names line and check the format against the format line
sub parse_line {
    my ($format, $name, $line) = @_;
    my %retval = ();

    # Load the format hash
    my %fmt = 
	("n" => '\d+',          # Simple number (eg 4)
	 "f" => '\d+(?:.\d+)?', # Floating point number (eg 4.5)
	 "s" => '\S+',          # Some text string
	 );

    # Split up the input
    my @formats = split /\s+/, $format;
    my @names   = split /\s+/, $name;
    my @values  = split /\s+/, $line;

    # Check the input is valid
    die "Bad format or names"
	unless int @formats == int @names;

    die "Bad line: $line"
	unless int @formats == int @values;

    # Read the data
    my ($f, $n, $v) = (pop @formats, pop @names, pop @values);
    while (defined $f and defined $v and defined $n) {

	die "Data does not match format: $f != $v"
	    unless $v =~ m|$fmt{$f}|;
	
	$retval{$n} = $v;

	($f, $n, $v) = (pop @formats, pop @names, pop @values);
    }
    
    # Return
    %retval;
}

# Make sure the given quad is valid
sub check_quad {
    my ($obj) = shift;
    my ($A, $B, $C, $D) = @{$obj->points};

    # Work out the line segments given the points
    my $AD = $D->subtract($A);
    my $AB = $B->subtract($A);
    my $AC = $C->subtract($A);
    my $BA = $A->subtract($B);
    my $BC = $C->subtract($B);
    my $CB = $B->subtract($C);
    my $CD = $D->subtract($C);
    my $DC = $C->subtract($D);
    my $DA = $A->subtract($D);

    # Compute the cross products that will be used later
    my $ADxAB = $AD->cross($AB);
    my $BAxBC = $BA->cross($BC);
    my $CBxCD = $CB->cross($CD);
    my $DCxDA = $DC->cross($DA);

    # Make sure the quad is planar (i.e. all points fall in the same plane)
    # LDraw quads do not need to be but in OpenGL they must be, so any 
    # non-planar quad will need to be tesselated
    # 
    # From Eben Scanlon:
    # The test (ADxAB) . AC will be the same as the determinant.
    # Where x is the cross product and . is the dot product
    my $d = $ADxAB->dot($AC);
    
    if ($d != 0) {
	print STDERR "Quad not planar in '", $obj->parent->filename.
	    "' line ", $obj->line_no, ":\n\t".
	    $obj->print, "\n";
	print STDERR "Determinant: $d\n";
	exit;
    }
    
    # From Eben Scanlon:
    # To checkboth whether the polygon intersects and if it is convex, all:
    # ADxAB, BAxBC, CBxCD, DCxDA
    # Must point in the same direction (have the same sign) so 1 normal
    # might be 1, -2, 5 and the rest must follow the sign format +, -, +.
    # Beware 0s.
    my (@dir) = (1,
		 same_direction($ADxAB, $BAxBC),
		 same_direction($ADxAB, $CBxCD),
		 same_direction($ADxAB, $DCxDA));
    
    # These are all normals to the surface so we pick one and
    # normalize it to a length of 1
    $obj->{"NORMAL"} = $ADxAB->normalize;

    # Count the number in the same direction
    my $count = sum(@dir);

    # Check to see if the polygon is convex or twisted
    unless ($count == 0 or $count == 4) {
	unless ($count == 2) {
	    print STDERR "Polygon is concave in '", $obj->parent->filename.
		"' line ", $obj->line_no, ":\n\t".
		    $obj->print, "\n";
	    print STDERR "\tNormal directions: ", join(":", @dir), "\n";
	    exit;
	}
	
	# If it is twisted, "untwist" it
	my $i = 0;
	for ($i = 0; $i < int @dir; $i++) {
	    my $j = ($i + 1) % int @dir;
	    
	    if ($dir[$i] == 0 and $dir[$j] == 0) {
		$dir[$i] = 1; $dir[$j] = 1;
		
		# Swap the points
		my $t = $obj->points->[$i];
		$obj->points->[$i] = $obj->points->[$j];
		$obj->points->[$j] = $t;
		
		# And fix the normal
		$obj->renormalize;
	    }
	}

    }

}

# Sum the arguments
sub sum {
    my $r;
    while ($_ = shift) {
	$r += $_;
    }

    return $r;
}

# Compare the direction of the given vectors
# Return 1 if they are the same direction, or 0 otherwise
sub same_direction {
    my ($a, $b) = @_;
    croak "Vectors were not given" unless ref $a and ref $b;

    # Make sure that the vectors are the same size
    croak "Vectors must be the same size" unless $a->N == $b->N;

    # Iterate through the vector and test to make sure the signs are the same
    my $i;
    foreach $i ($a->Nindices) {
	# Check each component against 0
	# <=> 0 returns 1 if > 0, 0 if 0, 1 if < 0
	# If the two <=> pieces are equal then both are on the same side of 0
	return 0 unless
	    ($a->get($i) <=> 0) == ($b->get($i) <=> 0);
    }
    
    return 1;
}

1;
