#!/usr/local/bin/perl -w

# This provides a class for reading in LDraw format files

package LDraw;

use strict;
use vars qw(@ISA @EXPORT @EXPORT_OK %EXPORT_TAGS $VERSION);

# Allow carping of errors
use Carp;

# Use the object oriented filehandle interface
use IO::File;

# Allow us to get absolute paths from filenames
use Cwd 'fast_abs_path';
use File::Basename;

# Include the LDObj routines
use LDObj;

use Exporter;
$VERSION = 1.00;
@ISA = qw(Exporter);

@EXPORT      = qw();
@EXPORT_OK   = qw();
%EXPORT_TAGS = ();


# This allows us to keep track of read files on a classwide basis
my %FILES = ();
my @FILE_NOS = ();
my @LD_PATH = ();     # This holds the LDraw file search path

=head1 NAME

LDraw - Read LDraw format files and provide a method for accessing the data

=head1 SYNOPSIS

    use LDraw;

    my $ld = LDraw->new("LDrawFile.txt", ['.']);

=cut

=head2 LDraw->new($filename);

This reads the specified file and parses it, building an object that
can be used to access the LDraw data in a useful way.

If the path argument is given it should be an arrayref of either
absolute or relative paths.  If the file given exists then the path to
the file is added to the front of the path.

This returns a LDraw object.

=cut

sub new {
    my $class = shift;
    my $file = shift 
	or croak "no LDraw filename specified";
    my $ld_path = shift;
    

    # Get the class name if we were called through an object
    $class = ref $class if ref $class;

    # Set up the path if needed
    if (defined $ld_path and int(@LD_PATH) == 0) {
	@LD_PATH = @$ld_path;

	my $p;
	foreach $p (@LD_PATH) {
	    $p = fast_abs_path($p);
	}
    }

    # Add the path to the given file to the path if the file exists
    if (-e $file) {
	my ($name, $path, $suffix) = fileparse($file, '\..*');
	$path = fast_abs_path($path);
	$file = $path . '/' . $name . $suffix;

	unshift @LD_PATH, $path
	    if int(@LD_PATH) == 0 or $LD_PATH[0] ne $path;
    }

    # Make sure we have a path
    croak "Need to specify a path to LDraw"
	unless int(@LD_PATH) != 0;

    # Check the search path for the file if the file does not exist
    my $p;
    foreach $p (@LD_PATH) {
	# Clean the filename, expand it and make it absolute
	my ($name, $path, $suffix) = fileparse($file, '\..*');

	$file = $p . '/' . $name . $suffix;

	last if -e $file;
    }
    
    croak "Unable to find file '$file' in path: '" .
	join(':', @LD_PATH), "'"
	    unless -e $file;


    # See if the file is defined and return it if it is
    if (defined $FILES{$file}) {
	if ($FILES{$file} != 1) {
	    return $FILES{$file};
	} else {
	    croak "Circular dependency in the data files";
	}
    }
    
    # Set up memory for the object and bless it into the class
    my $self = {};
    bless($self, $class);

    # Initialize the object
    $self->_init($file);

    # Store the file
    $self->{"NUM"} = int @FILE_NOS;
    $FILE_NOS[$self->number] = $self;
    $FILES{$file} = $self;

    # And give the object back to the caller
    return $self;
}


=head2 $ldraw->walk($actions);

This will traverse all of the objects of the specified type and 
run the action for that type.

$actions is a hashref for each type to work on.  Only the types
with a defined action will be walked so:
    $actions = {1 => \&check_obj,
		4 => \&check_quad,
	       };
Will only walk the quads and the objects.  Note that subobjects 
will be checked by default.  If you do not want this behavior 
explicitly send in an undef for type 1, or define your own
object checker instead.

The action subroutine will be passed each LDObj that is 
encountered, and is expected to return 1 if the check is successful 
0 if it fails and to set $! to the reason the check failed. 

This will either return 1 to indicate success.  On failure 0 is
returned and $! will indicate the first error that occured.

=cut

sub walk {
    my $self = shift;
    my $actions = shift;
    croak "instance method called on class" unless ref $self;
    croak "bad actions" unless ref $actions;

    # Walk all the types only if the action is defined
    my $type;
    for ($type = 0; $type < int @{$self->{"TYPES"}}; $type++) {
	if ($type == 1 or defined $actions->{$type})	    
	{
	    # Walk the tree
	    my $obj;
	    foreach $obj (@{$self->{"TYPES"}[$type]}) {
		# If the type is one and there is no type defined 
		if ($type == 1 and not exists $actions->{$type}) {

		    # Walk the child
		    $obj->subpart->walk($actions);
		}
		elsif (defined $actions->{$type}) {
		    &{$actions->{$type}}($obj, $actions);
		}
	    }
	}
    }

    return 1;
}


=head2 $ldraw->get_objs_by_type();

This returns the B<raw> list of the types as a arrayref keyed on the type
and the value for each is an arrayref to the LDObj list of all of the
objects of that type.  If you modify this list, do so at your own risk.

=cut
    
sub get_objs_by_type {
    return shift->{"TYPES"};
}


=head2 $ldraw->get_objs_by_lineno();

This returns the B<raw> list of the objects as an arrayref where each
key is the lineno and the value is an LDObj containing the object
found on that line.  If you modify this list, do so at your own risk.

=cut

sub get_objs_by_lineno {
    return shift->{"LINES"};
}


=head2 $ldraw->get_bounding_points();

This returns the two bounding points as a hashref. The two points are
the keys 'MAX' and 'MIN' and they point to vectors that define a
rectangle that will fit around the object.  The rectangle is aligned
with the axes.

=cut

sub get_bounding_points {
    return shift->{"BOUNDING"};
}


=head2 $ldraw->number();

This gets the arbitrary file number for use when selecting.  Do not
rely upon the sequence of the value for anything.

=cut

sub number {
    return shift->{"NUM"};
}


=head2 $ldraw->filename();

This gets the filename used to create the object.

=cut

sub name {
    return shift->{"NAME"};
}


=head2 $ldraw->get_file($id);

This returns the LDraw object associated with file number $id.

=cut

sub get_file {
    my ($self, $file) = @_;
    return $FILE_NOS[$file];
}


=head2 $ldraw->get_line($line_no);

This returns the LDObj associated with line number $line_no.

=cut

sub get_line {
    my ($self, $line_no) = @_;
    return $self->{"LINES"}[$line_no];
}


=head2 $ldraw->get_fileno_list($id);

This returns a referenct to the LDraw file number list.  Note that you
can break things if you change this list.

=cut

sub get_fileno_list {
    return \@FILE_NOS;
}


####
# PRIVATE functions, do not use unless you absolutely have to
#   I will not guarantee that the below will not change
####


# Initialize the matrix
# Should probably be called by subclasses
sub _init {
    my $self = shift;
    my $file = shift;
    croak "Bad arguments to the constructor"
	unless ref $self and $file;


    # Set up the instance variables
    $self->{"NAME"} = $file;      # Our filename
    $self->{"LINES"} = [];        # Index the LD objects by line number
    $self->{"TYPES"} = [];        # Index the LD objects by type number
    $self->{"NUM"} = [];          # Store the number of the file for selection
    $self->{"BOUNDING"} = {       # Provide the bounding points for the file
	"MAX" => undef,
	"MIN" => undef,
	};
    
    # Open the specified file
    my $ldfh = new IO::File "$file", "r";
    croak qq(Unable to open file "$file": $!)
	unless $ldfh;

    # Read the file and do the actions
    my $line;
    while (defined( $line = $ldfh->getline ) ) {
	chomp $line;
	my $line_no = $.;
	
	my $obj = LDObj->new($line, $line_no, $self);
	
	$self->{"LINES"}[$line_no] = $obj;
	push @{$self->{"TYPES"}[$obj->type]}, $obj
	    unless $obj->type < 0;
	
	# Check the bounding box data
	if (_in($obj->type, [2,3,4]) ) {
	    _update_bounding($self->{"BOUNDING"}, $obj);
	}
	elsif ($obj->type == 1) {
	    _update_part_bounding($self->{"BOUNDING"}, $obj);
	}
    }

    # Close the filehandle
    undef $ldfh;

    return $self;
}

sub _update_part_bounding {
    my ($bound, $obj) = @_;

    my $t = $obj->translation;
    my $p = $obj->points->[0];
    my $sp = $obj->subpart;

    # The bounding of a subpart is the subparts translated bounding
    # plus the position offset
    my $sp_max = $sp->get_bounding_points->{"MAX"}->add($p);
    my $sp_min = $sp->get_bounding_points->{"MIN"}->add($p);

    $sp_max = _translate($sp_max, $t);  # Apply the translation
    $sp_min = _translate($sp_min, $t);  # Apply the translation

    if (not defined $bound->{"MAX"}) {
	$bound->{"MAX"} = $sp_max->clone;
	$bound->{"MIN"} = $sp_min->clone;
    }

    my $max = $bound->{"MAX"};
    my $min = $bound->{"MIN"};

    my $i;
    foreach $i (0 .. 2) {	    
	$max->set($i, _max($max->get($i), 
			   $sp_max->get($i),
			   $sp_min->get($i))
		  );
	$min->set($i, _min($min->get($i),
			   $sp_max->get($i),
			   $sp_min->get($i))
		  );
    }
}

# D is the vector to translate, T is the translation
sub _translate {
    my ($d, $t) = @_;
    my $r = $d->clone;
    
    my $i;
    foreach $i (0 .. 2) {
	my $v = 0;
	my $j;
	foreach $j (0 .. 2) {
	    $v += $d->get($i) * $t->[$j]->get($i);
	}
	$r->set($i, $v);
    }

    return $r;
}

sub _update_bounding {
    my ($bound, $obj) = @_;

    if (not defined $bound->{"MAX"}) {
	$bound->{"MAX"} = $obj->points->[0]->clone;
	$bound->{"MIN"} = $obj->points->[0]->clone;
    }

    my $max = $bound->{"MAX"};
    my $min = $bound->{"MIN"};

    my ($p, $i);
    foreach $p (@{$obj->points}) {
	foreach $i (0 .. 2) {	    
	    $max->set($i, _max($max->get($i), $p->get($i)) );
	    $min->set($i, _min($min->get($i), $p->get($i)) );
	}
    }
}

sub _max {
    my $r = shift;
    my $v;
    foreach $v (@_) {
	$r = $v 
	    if $v > $r;
    }
    $r;
}

sub _min {
    my $r = shift;
    my $v;
    foreach $v (@_) {
	$r = $v 
	    if $v < $r;
    }
    $r;
}

sub _in {
    my ($val, $list) = @_;

    my $i = int(@$list) - 1;
    while ($i >= 0) {
	return 1
	    if $list->[$i] eq $val;

	$i--;
    }
    return 0;
}

1;
