#!/usr/local/bin/perl -w

# This provides a class for manipulating NxM matrices

package Matrix;

use strict;
use vars qw(@ISA @EXPORT @EXPORT_OK %EXPORT_TAGS $VERSION);

# Allow carping of errors
use Carp;

# Can overload with the overload package by uncommenting the following
#  and filling in the appropriate functions
# use overload ('+' => \&matrix_add);

use Exporter;
$VERSION = 1.00;
@ISA = qw(Exporter);

@EXPORT      = qw();
@EXPORT_OK   = qw();
%EXPORT_TAGS = ();


=head1 NAME

Matrix - Store and manipulate an MxN matrix

=head1 SYNOPSIS

    M is the number of columns (or X)
    N is the number of rows (or Y)
    
    my $matrix =  Matrix->new("-n" => 3, "-m" => 2,
			      -data => [[1, 2],
					[2, 5],
					[0, 6]]
			      );
    print $matrix->print;

    This should print:
        1  2
        2  5
        0  6

=cut

=head2 Matrix->new(%ARGS);

This accepts arguments in hash form:

   Matrix->new("-n" => 3, "-m" => 2)

Which makes a new 2x3 matrix with all elements zeroed.


The valid argument names are:

=over 4

=item C<-n>: N dimension (rows)

=item C<-m>: M dimension (columns)

=item C<-data>: [[] .. []] data information in arrayrefs of arrayrefs
if the data is larger than C<-n> & C<-m> then it is cut off,
smaller data is zeroed

=item C<-nocopy>: Copy the data that is passed to us.

=back

The leading - is optional on args but recommended.  The C<-n> and C<-m>
are optional if data is specified in which case it will use the -data arg
to determine the size of the matrix.  The matrix will be created with 
the maximum possible size and zeroes fill in missing data.

Note that it is never okay to create a matrix without specifying the size
in some way, either explicitly or giving data to infer the size from.

This returns a Matrix object.

=cut

sub new {
    my $class = shift;
    
    # Get the class name if we were called through an object
    $class = ref $class if ref $class;

    # Set up memory for the object and bless it into the class
    my $self = {};
    bless($self, $class);

    # Initialize the object with the remaining arguments
    $self->_init(@_);
    
    # And give the object back to the caller
    return $self;
}

sub clone {
    my $self = shift;
    croak "instance method called on class" unless ref $self;

    # Get the right args from the existing object and make a new one
    return $self->new(-m => $self->M,
		      -n => $self->N,
		      -nocopy => $self->{"DATA"},
		      );
}


=head2 $matrix->M; $matrix->N;

These return the dimensions of the matrix.  The dimensions can not 
change after creation.

=cut

sub M {
    my $self = shift;
    croak "instance method called on class" unless ref $self;
    return $self->{"M"};
}
sub N {
    my $self = shift;
    croak "instance method called on class" unless ref $self;
    return $self->{"N"};
}


=head2 $matrix->Mindices; $matrix->Nindices;

These return the all of the valid M or N indices so either 0 .. M or
0 .. N.

=cut

sub Mindices {
    my $self = shift;
    croak "instance method called on class" unless ref $self;
    return (0 .. ($self->{"M"}-1) );
}
sub Nindices {
    my $self = shift;
    croak "instance method called on class" unless ref $self;
    return (0 .. ($self->{"N"}-1) );
}


=head2 $matrix->get($m, $n);
=head2 $matrix->set($m, $n, $value);

This will either get the value of matrix entry $m, $n.
Or it will set matrix entry $m, $n to $value and return the new value.

The first element in a matrix is 0, 0.  Note that N and M are never inside
the matrix, the last element is N-1, M-1.

=cut

sub get {
    my $self = shift;

    # If we were called through the class then we need to shift 
    # again to get the object
    $self = shift
	unless ref $self;

    my $m = shift;
    my $n = shift;
    croak "insufficient arguments" 
	unless defined($m) and defined($n);

    # Check the sizes
    croak "attempt to access an element outside the matrix"
	unless ($m < $self->M and $n < $self->N);

    # Call the function that does the work
    return $self->_get($m, $n);
}

sub set {
    my $self = shift;

    # If we were called through the class then we need to shift 
    # again to get the object
    $self = shift
	unless ref $self;

    my $m = shift;
    my $n = shift;
    my $value = shift;
    croak "insufficient arguments" 
	unless defined($m) and defined($n) and defined($value);

    # Check the sizes
    croak "attempt to access an element outside the matrix"
	unless ($m < $self->M and $n < $self->N);

    # Call the function that does the work
    return $self->_set($m, $n, $value);
}


=head2 $matrix->determinant;

This computes the determinant of the matrix.  This can be a costly 
operation on large matrices.

Also the matrix given must be square.

The return value is the numerical determinant.

=cut

sub determinant {
    my $self = shift;
    croak "instance method called on class" unless ref $self;

    # Check the matrix dimensions
    croak "Unable to compute determinant on non-square matrix"
	unless $self->M == $self->N;

    return det($self->{"DATA"});
}


=head2 $matrix1->add($matrix2)

This computes the result of $matrix1 + $matrix2.

Matrices to be added must be the same size.

The return value is a new matrix with the sum of the two input matrices.

=cut

sub add {
    my $self = shift;
    my $other = shift;
    croak "instance method called on class" unless ref $self;
    return $self->iterate($other, sub { my $a = shift; $a += shift; $a } );
}


=head2 $matrix1->subtract($matrix2)

This computes the result of $matrix1 - $matrix2.

Matrices to be subtracted must be the same size.

The return value is a new matrix with the result of the two input matrices.

=cut

sub subtract {
    my $self = shift;
    my $other = shift;
    croak "instance method called on class" unless ref $self;
    return $self->iterate($other, sub { my $a = shift; $a -= shift; $a } );
}


=head2 $matrix1->multiply($matrix2)

This multiplies $matrix1 by $matrix2 and returns a new matrix 
with the result.

Matrices to be multiplied must have the following property:
    $matrix1->M == $matrix2->N

The return value is a new matrix with the product of the two input matrices.
Its size will be $matrix1->N by $matrix2->N.

=cut

sub multiply {
    my $self = shift;
    my $other = shift;
    croak "instance method called on class" unless ref $self;
    croak "no second matrix supplied" unless ref $other;

    # Check the matrix dimensions
    croak "Unable to compute the product unless the colums of 1 " .
	"equals the rows of 2"
	    unless $self->M == $other->N;

    # Create the new matrix
    my $result = Matrix->new("-N" => $self->N, "-M" => $other->M);

    # Calculate the product
    my ($a, $b, $d) = (0, 0, 0);
    my $oM = $other->M - 1;
    my $sN = $self->N - 1;
    my $sM = $self->M - 1;
    foreach $d (0 .. $oM) {
	foreach $b (0 .. $sN) {
	    my $r = 0;
	    foreach $a (0 .. $sM) {
		$r += $self->_get($a, $b) * $other->_get($d, $a);
	    }
	    $result->_set($d,$b, $r)
	}
    }

    return $result;
}


=head2 $matrix->name

This returns the name of the object as a string.

=cut

sub name {
    my $self = shift;
    croak "instance method called on class" unless ref $self;
    
    return $self->{"NAME"};
}


=head2 $matrix->print

This returns the contents of the matrix with some attempt to pretty 
print it.  The format is not yet fixed.

=cut

sub print {
    my $self = shift;

    # If we were called through the class then we need to shift 
    # again to get the object
    $self = shift
	unless ref $self;

    my $ret = "";

    $ret .= $self->name . " is " . $self->M . " x " . $self->N . "\n\n";

    my ($m, $n) = (0, 0);
    my $sN = $self->N - 1;
    my $sM = $self->M - 1;
    foreach $n (0 .. $sN) {
	foreach $m (0 .. $sM) {
	    $ret .= Matrix->_get($self,$m,$n) . "\t";
	}
	$ret .= "\n";
    }

    return $ret;
}

####
# PRIVATE functions, do not use unless you absolutely have to
#   I will not guarantee that the below will not change
####

# Initialize the matrix
# Should probably be called by subclasses
sub _init {
    my $self = shift;
    
    # Set up the base arguments
    $self->{"N"} = undef;
    $self->{"M"} = undef;
    $self->{"DATA"} = undef;
    $self->{"NAME"} = undef;
    $self->{"ONE_DIM"} = 0;  # 0 - multi dim, 1 = M, -1 = N

    # Examine the arguments
    # Did we get any arguments?
    croak "No size was specified to the matrix constructor"
	if @_ == 0;

    # Are we not allowed to copy the data?
    my $nocopy = 0;

    # Iterate through the arguments and populate the data
    my $arg;
    while (defined ($arg = shift) ) {
	$arg = uc $arg;
	$arg =~ s|\s||g;
	$arg =~ s|^-||;

	# Set the nocopy flag
	if ($arg eq "NOCOPY") {
	    $nocopy = 1;
	    $arg = "DATA";
	}

	if (exists $self->{$arg}) {
	    carp "Defining $arg twice"
		if defined $self->{$arg};
	    $self->{$arg} = shift;
	} else {
	    carp "Bad argument $arg";
	}
    }

    # Check that we got what we needed and try to fill in anything that 
    # is missing
    # Check the name
    $self->{"NAME"} = $self->{"NAME"} || "Matrix";
    
    # If N & M was not given then look at the data to determine the size
    $self->{"N"} = $self->{"N"} || @{$self->{"DATA"}};

    if (not defined $self->{"M"}) {
	my $m = 0;
	my $row;
	for ($row = 0; $row < $self->{"N"}; $row++) {
	    $m = max($m, int @{$self->{"DATA"}[$row]});
	}

	$self->{"M"} = $m;
    }
    carp "M & N must be greater than zero"
	unless ($self->M > 0 and $self->N > 0);

    # If there is only one row or column defined and we were passed 
    # an array ref then store it as an array rather than a 2D matrix
    if (($self->M == 1 or $self->N == 1) and 
	(not defined $self->{"DATA"} or not ref $self->{"DATA"}->[0])
	)
    {
	# Define data if it was not specified
	$self->{"DATA"} = []
	    unless defined $self->{"DATA"};

	# Set this to the right dimension
	$self->{"ONE_DIM"} = $self->M != 1 
	    ? 1   # M
	    : -1; # N

	# Copy the data into our internal array if we have to
	if ($nocopy) {
	    $self->{"DATA"} = [@{$self->{"DATA"}}];
	}
    }
    
    else {
	# Copy the data into our internal array if we have to
	if ($nocopy) {
	    my $data = $self->{"DATA"};
	    $self->{"DATA"} = ();
	    
	    my ($m, $n) = (0, 0);
	    my $sM = $self->M - 1;
	    my $sN = $self->N - 1;
	    foreach $m (0 .. $sM) {
		foreach $n (0 .. $sN) {
		    $self->{"DATA"}[$n][$m] = def_or_zero($data->[$n][$m]);
		}
	    }
	}
    }

    return $self;
}

# Returns the argument if it is defined, otherwise returns 0
sub def_or_zero {
    return $_[0] if defined $_[0];
    return 0;
}

# Returns the largest of the arguments
sub max {
    my $l = shift;
    my $c = 0;
    while (defined ($c = shift)) {
	$l = $c 
	    if $c > $l;
    }

    return $l;
}


# Compute the determinant
# The single argument is an arrayref of arrayrefs pointing to the data
# It returns a scalar which is the determinant
sub det {
    my $data = shift;

    # Work out the size
    my $n = int @$data;

    # Matrix where N = 1
    if ($n == 1) {
	return $data->[0][0];
    }

    # Matrix where N = 2
    if ($data == 2) {
	return ($data->[0][0] * $data->[1][1]) - 
	    ($data->[1][0] * $data->[0][1]);
    }
    
    # Matrix where N > 2
    my $r = 0;
    my $i = 0;
    my $j;
    for ($j = 0; $j < $n; $j++) {
	$r += (($i + $j) % 2 ? -1 : 1 )  # Simulates -1^($i+$j)
	    * $data->[$i][$j]            # Element $i $j of matrix $m
	    * det(minor($data, $i, $j))  # Determinant of the minor
	    ;
    }

    # Return
    $r;
}

# Return an array ref of the minor of the given 2 dimensional array ref
# A minor is the portion of the matrix except the row and column given 
#  by $i and $j respectively
sub minor {
    my ($m, $i, $j) = @_;
    my $M = ();

    # Copy the matrix elements into the minor
    my $o = -1;
    my $q;
    for ($q = 0; $q < int @$m; $q++) {
	next if $q == $j;
	$o++;

	my $n = -1;
	my $p;
	for ($p = 0; $p < int @$m; $p++) {
	    next if $p == $i;
	    $n++;

	    $M->[$n][$o] = $m->[$p][$q];
	}
    }

    # Return
    $M;
}

# Iterate over a pair of matrices and apply the operation to both
sub iterate {
    my $self = shift;
    my $other = shift;
    my $operation = shift;
    croak "instance method called on class" unless ref $self;
    croak "no second matrix supplied" unless ref $other;
    croak "no operation specified" unless ref $operation;

    # Check the matrix dimensions
    croak "Unable to apply operation to different sized matrices"
	unless $self->M == $other->M and 
	    $self->N == $other->N;

    # Create the new matrix
    my $result = $self->new("-M" => $self->M, "-N" => $self->N);

    # Do the operation
    my ($m, $n) = (0, 0);
    my $sN = $self->N - 1;
    my $sM = $self->M - 1;
    foreach $m (0 .. $sM) {
	foreach $n (0 .. $sN) {
	    my $value = &$operation($self->_get($m,$n), $other->_get($m,$n));
	    $result->_set($m,$n, $value);
	}
    }

    return $result;
}


sub _get {
    my $self = shift;

    # If we were called through the class then we need to shift 
    # again to get the object
    $self = shift
	unless ref $self;

    # Get the arguments
    my $m = shift;
    my $n = shift;

    # Get the data
    my $ret = undef;
    if ($self->{"ONE_DIM"} > 0) {
	$ret = $self->{"DATA"}[$m];
    } elsif ($self->{"ONE_DIM"} < 0) {
	$ret = $self->{"DATA"}[$n];
    } else {
	$ret = $self->{"DATA"}[$n][$m];
    }

    # Return 0 if array not defined at the point given
    def_or_zero($ret);    
}

sub _set {
    my $self = shift;

    # If we were called through the class then we need to shift 
    # again to get the object
    $self = shift
	unless ref $self;

    my $m = shift;
    my $n = shift;
    croak "insufficient arguments" 
	unless defined($m) and defined($n);

    # Check the sizes
    croak "attempt to access an element outside the matrix"
	unless ($m < $self->M and $n < $self->N);

    # Get the value
    my $value = shift;
    croak "too many arguments" if @_;
    
    # Set the value
    if ($self->{"ONE_DIM"} > 0) {
	return $self->{"DATA"}[$m] = $value;
    } elsif ($self->{"ONE_DIM"} < 0) {
	return $self->{"DATA"}[$n] = $value;
    }
    return $self->{"DATA"}[$n][$m] = $value;
}

1;
