#!/usr/local/bin/perl -w

# This provides a class for manipulating Quaternions
# Stolen from "OpenGL Programming for the X Window System" by Mark Kilgard

package Quat;

use strict;
use vars qw(@ISA @EXPORT @EXPORT_OK %EXPORT_TAGS $VERSION);

# Allow carping of errors
use Carp;

use Exporter;
$VERSION = 1.00;
@ISA = qw(Exporter);

@EXPORT      = qw(trackball add_quats build_rotmatrix axis_to_quat);
@EXPORT_OK   = qw();
%EXPORT_TAGS = ();


# This size should really be based on the distance from the
# center of rotation to the point on the object underneath the
# mouse.  That point would then track the mouse as closely as
# possible.  This is a simple example, though, so that is left
# as an Exercise for the Programmer.
sub TRACKBALLSIZE  { 10; }

# void vzero(float *v)
sub vzero {
    my $v = shift;

    $v->[0] = 0.0;
    $v->[1] = 0.0;
    $v->[2] = 0.0;
}

# void vset(float *v, float x, float y, float z)
sub vset {
    my ($v, $x, $y, $z) = @_;

    $v->[0] = $x;
    $v->[1] = $y;
    $v->[2] = $z;
}

# void vsub(const float *src1, const float *src2, float *dst)
sub vsub {
    my ($src1, $src2, $dst) = @_;

    $dst->[0] = $src1->[0] - $src2->[0];
    $dst->[1] = $src1->[1] - $src2->[1];
    $dst->[2] = $src1->[2] - $src2->[2];
}

# void vcopy(const float *v1, float *v2)
sub vcopy {
    my ($v1, $v2) = @_;
    
    my $i;
    for ($i = 0; $i < 3; $i++) {
	$v2->[$i] = $v1->[$i];
    }
}

# void vcross(const float *v1, const float *v2, float *cross)
sub vcross {
    my ($v1, $v2, $cross) = @_;

    my @temp;
    $temp[0] = ($v1->[1] * $v2->[2]) - ($v1->[2] * $v2->[1]);
    $temp[1] = ($v1->[2] * $v2->[0]) - ($v1->[0] * $v2->[2]);
    $temp[2] = ($v1->[0] * $v2->[1]) - ($v1->[1] * $v2->[0]);
    vcopy(\@temp, $cross);
}

# float vlength(const float *v)
sub vlength {
    my ($v) = @_;
    
    return sqrt($v->[0] * $v->[0] + $v->[1] * $v->[1] + $v->[2] * $v->[2]);
}

# void vscale(float *v, float div)
sub vscale {
    my ($v, $div) = @_;

    $v->[0] *= $div;
    $v->[1] *= $div;
    $v->[2] *= $div;
}

# void vnormal(float *v)
sub vnormal {
    my ($v) = @_;

    vscale($v, 1.0 / vlength($v));
}

# float vdot(const float *v1, const float *v2)
sub vdot {
    my ($v1, $v2) = @_;

    return $v1->[0] * $v2->[0] + $v1->[1] * $v2->[1] + $v1->[2] * $v2->[2];
}

# void vadd(const float *src1, const float *src2, float *dst)
sub vadd {
    my ($src1, $src2, $dst) = @_;

    $dst->[0] = $src1->[0] + $src2->[0];
    $dst->[1] = $src1->[1] + $src2->[1];
    $dst->[2] = $src1->[2] + $src2->[2];
}

# void trackball(float q[4], float p1x, float p1y, float p2x, float p2y)
sub trackball {
    my ($q, $p1x, $p1y, $p2x, $p2y) = @_;

    my $a = [];           # Axis of rotation. 
    my $phi = 0.0;        # How much to rotate about axis.
    my ($p1, $p2, $d) = ([], [], []);
    my $t = 0.0;

    if ($p1x == $p2x and $p1y == $p2y) {
	# Zero rotation 
	vzero($q);
	$q->[3] = 1.0;
	return;
    }

    # First, figure out z-coordinates for projection of P1 and
    # P2 to deformed sphere. 
    vset($p1, $p1x, $p1y, tb_project_to_sphere(TRACKBALLSIZE, $p1x, $p1y));
    vset($p2, $p2x, $p2y, tb_project_to_sphere(TRACKBALLSIZE, $p2x, $p2y));

    # Now, we want the cross product of P1 and P2.
    vcross($p2, $p1, $a);

    # Figure out how much to rotate around that axis.
    vsub($p1, $p2, $d);
    $t = vlength($d) / (2.0 * TRACKBALLSIZE);

    # Avoid problems with out-of-control values.
    $t = 1.0
	if ($t > 1.0);
    $t = -1.0
	if ($t < -1.0);

    $phi = 2.0 * asin($t);
    
    axis_to_quat($a, $phi, $q);
}

sub asin {
    atan2($_[0], sqrt(1 - $_[0] * $_[0]));
}

# Given an axis and angle, compute quaternion.
# void axis_to_quat(float a[3], float phi, float q[4])
sub axis_to_quat {
    my ($a, $phi, $q) = @_;

    vnormal($a);
    vcopy($a, $q);
    vscale($q, sin($phi / 2.0));
    $q->[3] = cos($phi / 2.0);
}

# Project an x,y pair onto a sphere of radius r OR a
# hyperbolic sheet if we are away from the center of the
# sphere.
# static float tb_project_to_sphere(float r, float x, float y)
sub tb_project_to_sphere {
    my ($r, $x, $y) = @_;
    
    my ($d, $t, $z) = (0, 0, 0);

    $d = sqrt($x * $x + $y * $y);
    if ($d < $r * 0.70710678118654752440) {  # Inside sphere.
	$z = sqrt($r * $r - $d * $d);
    } else {              # On hyperbola.
	$t = $r / 1.41421356237309504880;
	$z = $t * $t / $d;
    }
    return $z;
}

# Given two rotations, e1 and e2, expressed as quaternion
# rotations, figure out the equivalent single rotation and
# stuff it into dest.  This routine also normalizes the result
# every RENORMCOUNT times it is called, to keep error from
# creeping in.  NOTE: This routine is written so that q1 or q2
# may be the same as dest (or each other).

sub RENORMCOUNT { 97; }

# void add_quats(float q1[4], float q2[4], float dest[4])
sub add_quats {
    my ($q1, $q2, $dest) = @_;

    my $count = 0;
    my ($t1, $t2, $t3) = ([], [], []);
    my $tf = [];
    
    vcopy($q1, $t1);
    vscale($t1, $q2->[3]);

    vcopy($q2, $t2);
    vscale($t2, $q1->[3]);

    vcross($q2, $q1, $t3);
    vadd($t1, $t2, $tf);
    vadd($t3, $tf, $tf);
    $tf->[3] = $q1->[3] * $q2->[3] - vdot($q1, $q2);
    
    $dest->[0] = $tf->[0];
    $dest->[1] = $tf->[1];
    $dest->[2] = $tf->[2];
    $dest->[3] = $tf->[3];
    
    if (++$count > RENORMCOUNT) {
	$count = 0;
	normalize_quat($dest);
    }
}

# Quaternions always obey:  a^2 + b^2 + c^2 + d^2 = 1.0 If
# they don't add up to 1.0, dividing by their magnitude will
# renormalize them.
# static void normalize_quat(float q[4])
sub normalize_quat {
    my ($q) = shift;

    my $i = 0;
    my $mag = 0;

    $mag = ($q->[0] * $q->[0] + $q->[1] * $q->[1] + 
	    $q->[2] * $q->[2] + $q->[3] * $q->[3]);
    for ($i = 0; $i < 4; $i++) {
	$q->[$i] /= $mag;
    }
}

# Build a rotation matrix, given a quaternion rotation.
# void build_rotmatrix(GLfloat m[4][4], float q[4])
sub build_rotmatrix {
    my ($m, $q) = @_;
    
    $m->[0][0] = 1.0 - 2.0 * ($q->[1] * $q->[1] + $q->[2] * $q->[2]);
    $m->[0][1] = 2.0 * ($q->[0] * $q->[1] - $q->[2] * $q->[3]);
    $m->[0][2] = 2.0 * ($q->[2] * $q->[0] + $q->[1] * $q->[3]);
    $m->[0][3] = 0.0;
    
    $m->[1][0] = 2.0 * ($q->[0] * $q->[1] + $q->[2] * $q->[3]);
    $m->[1][1] = 1.0 - 2.0 * ($q->[2] * $q->[2] + $q->[0] * $q->[0]);
    $m->[1][2] = 2.0 * ($q->[1] * $q->[2] - $q->[0] * $q->[3]);
    $m->[1][3] = 0.0;
    
    $m->[2][0] = 2.0 * ($q->[2] * $q->[0] - $q->[1] * $q->[3]);
    $m->[2][1] = 2.0 * ($q->[1] * $q->[2] + $q->[0] * $q->[3]);
    $m->[2][2] = 1.0 - 2.0 * ($q->[1] * $q->[1] + $q->[0] * $q->[0]);
    $m->[2][3] = 0.0;
    
    $m->[3][0] = 0.0;
    $m->[3][1] = 0.0;
    $m->[3][2] = 0.0;
    $m->[3][3] = 1.0;
}

1;
