#!/usr/local/bin/perl -w

# This provides a class for displaying simple LDraw shapes in OpenGL

package Render;

use strict;
use vars qw(@ISA @EXPORT @EXPORT_OK %EXPORT_TAGS $VERSION);

# Allow carping of errors
use Carp;

# Allow access to the OpenGL 3D graphics routines
use OpenGL;

# We manipulate LDObjs
use LDObj;

# We manipulate LDraws
use LDraw;

# And vectors
use Vector;

use Exporter;
$VERSION = 1.00;
@ISA = qw(Exporter);

@EXPORT      = qw(translate draw_poly draw_line check_for_glerror
		  walk_all_display_lists set_color processhits
		  gl_init);
@EXPORT_OK   = qw();
%EXPORT_TAGS = ();


# Store the colors that we will use when traversing the objects
# The first element is the current color so unshift and shift the stack
my @color_stack = (16);

# Store the display lists for the files
# Keyed on filename, value is display list #
my @display_list = ();

# Are we using GL lighting? 1 is yes, 0 is no
my $lighting = 0;


sub gl_init {
    my ($widget, $height, $width) = @_;
	

    # Get the window id to draw into
    my $id = $widget->WindowId;

    # Init Opengl
    glpOpenWindow('attributes' => [GLX_RGBA, GLX_DOUBLEBUFFER,
				   GLX_DEPTH_SIZE, 1],
		  'width' => $width,
		  'height' => $height,
		  'parent' => $$id,
		  );

    glClearColor(1,1,1,1);
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

    if ($lighting) {
	# Set up lighting
	glLightModeli(GL_LIGHT_MODEL_LOCAL_VIEWER, 1);
	glLightfv(GL_LIGHT0, GL_POSITION, pack("f4", 20, 20, 20, 1));
	glLightfv(GL_LIGHT0, GL_DIFFUSE, pack("f4", 0.6, 0.6, 0.4, 1));
	glLightf(GL_LIGHT0, GL_CONSTANT_ATTENUATION, 0.1);
	glLightf(GL_LIGHT0, GL_LINEAR_ATTENUATION, 0.05);
	glLightfv(GL_LIGHT1, GL_POSITION, pack("f4", -20, -20, -20, 0));
	glLightfv(GL_LIGHT1, GL_DIFFUSE, pack("f4", 0.8, 1, 0.8, 1));
	
	# And set up the material information
	glMaterialfv(GL_FRONT_AND_BACK, GL_AMBIENT, pack("f3", 0.5, 0.5, 0.5));
	glMaterialfv(GL_FRONT_AND_BACK, GL_DIFFUSE, pack("f3", 0.7, 0.7, 0.5));
	glMaterialfv(GL_FRONT_AND_BACK, GL_SPECULAR,pack("f3", 0.9, 0.9, 0.9));
	glMaterialf(GL_FRONT_AND_BACK, GL_SHININESS, 32);
	glColorMaterial(GL_FRONT, GL_AMBIENT_AND_DIFFUSE);
    
	# Set the GL options
	#glEnable(GL_LIGHT0);
	#glEnable(GL_LIGHT1);
	#glEnable(GL_LIGHTING);
	#glEnable(GL_NORMALIZE);
	#glShadeModel(GL_SMOOTH);
	#glEnable(GL_CULL_FACE);
	#glCullFace(GL_BACK);
	#glFrontFace(GL_CCW);
    }
    
    # Set the GL options
    glEnable(GL_DEPTH_TEST);
    glDepthFunc(GL_LESS);
    glLineWidth(1);

    glEnable(GL_POLYGON_OFFSET_FILL);
    glPolygonOffset(1.0, 1.0);

    glMatrixMode(GL_PROJECTION);
    glPushMatrix;
    glLoadIdentity;
    glMatrixMode(GL_MODELVIEW);
    glLoadIdentity;
}


# Process the hits that come from a select operation and return
# the file# and line# of the closest object
sub processhits {
    my $hits= shift;

    return undef if $hits == 0;

    # Store the closest object
    my $min = undef;
    my $min_obj = undef;

    # Examine all the elements identified by the selection
    my $i;
    for ($i = 0; $i < $hits; $i++) {
	my $names = shift;
	my $z1 = shift;
	my $z2 = shift;

	# Foreach name in the stack, store the name
	my @name = ();
	my $j;
	for ($j = 0; $j < $names; $j++) {
	    push @name, shift;
	}

	# Is this a new closest element?
	if (not defined $min or $z1 < $min) {
	    $min = $z1;
	    $min_obj = \@name;
	}
    }

    # Get the last 2 elements
    my $last = int(@$min_obj) - 1;
    my ($l, $f) = ($min_obj->[$last], $min_obj->[$last-1]);
    
    return ($f, $l);
}

# Walk all of the display lists that exist
sub walk_all_display_lists {
    my ($ld) = @_;

    # Work out how many files we have to deal with
    my $filenos = $ld->get_fileno_list();
    my $size = int(@$filenos);

    # Make the display lists
    my $first = glGenLists($size);
    die "Unable to get $size displaylists from GL\n"
	if $size == 0;
    
    # And walk the lists and populate the display_list array
    my $i;
    for ($i = 0; $i < $size; $i++) {
	$display_list[$i] = $first + $i;

	walk_display_list($ld, $i);
    }

    return $display_list[$ld->number];
}

# Walk one file
sub walk_display_list {
    my ($ld, $file_no) = @_;

    # Define the display list (but do not draw it yet)
    glNewList($display_list[$file_no],GL_COMPILE);

    # Draw the object by walking it with drawing actions
    my $actions = {1 => \&translate,
		   2 => \&draw_line,
		   3 => \&draw_poly,
		   4 => \&draw_poly,
	       };
    $ld->get_file($file_no)->walk($actions);

    # Finish the display list
    glEndList();

}

# Translate a subpart in OpenGL
sub translate {
    my ($obj, $actions) = @_;


    # Get the translation and position information
    my $p = $obj->points->[0];
    my $t = $obj->translation;
    
    # Store the current color
    unshift @color_stack, $obj->color;

    # Store the matrix
    glPushMatrix;

    # Store the file number as a name
    glLoadName($obj->line_no);
    glPushName($obj->subpart->number);
    glPushName(0);
    
    # Update the translation matrix
    # | X'|   | X |     | a b c x |
    # | Y'| = | Y |  x  | d e f y |
    # | Z'|   | Z |     | g h i z |
    # | 1 |   | 1 |     | 0 0 0 1 |
    # The function call order goes down then left
    glMultMatrixf(pack("f16",
		       $t->[0]->get(0), $t->[1]->get(0), $t->[2]->get(0), 0,
		       $t->[0]->get(1), $t->[1]->get(1), $t->[2]->get(1), 0,
		       $t->[0]->get(2), $t->[1]->get(2), $t->[2]->get(2), 0,
		       $p->data_asrow, 1)
		  );
    
    if (0) {
	print "Matrix: ", 
	join(":","", 
	     $t->[0]->get(0), $t->[1]->get(0), $t->[2]->get(0), 0, "\n\t",
	     $t->[0]->get(1), $t->[1]->get(1), $t->[2]->get(1), 0, "\n\t",
	     $t->[0]->get(2), $t->[1]->get(2), $t->[2]->get(2), 0, "\n\t",
	     $p->data_asrow, 1, "\n"), "\n";
    }
    
    # Continue the walk
    if (defined $display_list[$obj->subpart->number]) {
	glCallList($display_list[$obj->subpart->number]);
    } else {
	print STDERR "No display list defined for ", $obj->subpart->number .
	    " drawing anyway but not using display list\n";
  
	$obj->subpart->walk($actions);
    }

    # And remove the current color
    shift @color_stack;

    # Restore the translation matrix
    glPopMatrix;

    # Restore the name stack
    glPopName;
    glPopName;
}


# Draw a poly in OpenGL
sub draw_poly {
    my $obj = shift;

    carp "Not a polygon"
	unless $obj->type == 3 or $obj->type == 4;

    # Identify the object
    glLoadName($obj->line_no);

    glBegin(GL_POLYGON); 
    {
	my @n = $obj->normal->data_asrow;
	#print "N: ", join(":", @n), "\n";

	glNormal3f( @n );
	glColor4f( convert_color($obj->color, $color_stack[0]) );
	
	my $p;
	foreach $p (@{$obj->points}) {
	    glVertex3f( $p->data_asrow );
	}
    }
    glEnd();
}

# Draw a line in OpenGL
sub draw_line {
    my $obj = shift;
    my ($A, $B) = @{$obj->points};

    # Identify the object
    glLoadName($obj->line_no);

    glBegin(GL_LINES); 
    {
	glColor4f( convert_color($obj->color, $color_stack[0]) );
	glVertex3f( $A->data_asrow );
	glVertex3f( $B->data_asrow );
    }
    glEnd();
}

# Set the color of the specified file number and part
sub set_color {
    my ($ld, $f, $l, $color) = @_;

    $ld->get_file($f)->get_line($l)->set_color(4);

    walk_display_list($ld, $f);
}

# Check to see if we got an error
sub check_for_glerror {
    my $error;
    
    while (($error = glGetError()) != GL_NO_ERROR) {
	print STDERR "GL Error: $error\n";
    }
}

# Store the color values
my @colors = 
    (
     #Color          Inverse Color RGBA            Dither RGBA
     ["black",             8,[0x22,0x22,0x22,0xff],[0x22,0x22,0x22,0xff]], # 0
     ["blue",              9,[0x00,0x33,0xb2,0xff],[0x00,0x33,0xb2,0xff]], # 1
     ["green",            10,[0x00,0x7f,0x33,0xff],[0x00,0x7f,0x33,0xff]], # 2
     ["teal",             11,[0x00,0xaa,0xaa,0xff],[0x00,0xaa,0xaa,0xff]], # 3
     ["red",              12,[0xcc,0x00,0x00,0xff],[0xcc,0x00,0x00,0xff]], # 4
     ["dk_pink",          13,[0xff,0x33,0x99,0xff],[0xff,0x33,0x99,0xff]], # 5
     ["brown",             0,[0x66,0x33,0x00,0xff],[0x66,0x33,0x00,0xff]], # 6
     ["grey",              8,[0xaa,0xaa,0xaa,0xff],[0xaa,0xaa,0xaa,0xff]], # 7
     ["dk_grey",           0,[0x66,0x66,0x58,0xff],[0x66,0x66,0x58,0xff]], # 8
     ["lt_blue",           1,[0x00,0x80,0xff,0xff],[0x00,0x80,0xff,0xff]], # 9
     ["lt_green",          2,[0x33,0xff,0x66,0xff],[0x33,0xff,0x66,0xff]], # 10
     ["cyan",              3,[0x55,0xaa,0xff,0xff],[0x55,0xaa,0xff,0xff]], # 11
     ["lt_red",            4,[0xff,0x55,0x55,0xff],[0xff,0x55,0x55,0xff]], # 12
     ["pink",              5,[0xff,0xb0,0xcc,0xff],[0xff,0xb0,0xcc,0xff]], # 13
     ["yellow",            8,[0xff,0xe5,0x00,0xff],[0xff,0xe5,0x00,0xff]], # 14
     ["white",             8,[0xff,0xff,0xff,0xff],[0xff,0xff,0xff,0xff]], # 15
     ["main_color",        0,[0x7f,0x7f,0x7f,0xff],[0x7f,0x7f,0x7f,0xff]], # 16
     ["pastel_green",      2,[0x66,0xf0,0x99,0xff],[0x66,0xf0,0x99,0xff]], # 17
     ["lt_yellow",        14,[0xff,0xff,0x80,0xff],[0xff,0xff,0x80,0xff]], # 18
     ["tan",               6,[0xcc,0xaa,0x66,0xff],[0xcc,0xaa,0x66,0xff]], # 19
     ["lt_purple",         8,[0xe0,0xcc,0xf0,0xff],[0xe0,0xcc,0xf0,0xff]], # 20
     ["glow_in_the_dark",  0,[0xe0,0xff,0xb0,0xff],[0xe0,0xff,0xb0,0xff]], # 21
     ["purple",            8,[0x99,0x33,0x99,0xff],[0x99,0x33,0x99,0xff]], # 22
     ["purple_blue",       8,[0x4c,0x00,0xcc,0xff],[0x4c,0x00,0xcc,0xff]], # 23
     ["edge_color",        0,[0x7f,0x7f,0x7f,0xff],[0x7f,0x7f,0x7f,0xff]], # 24
     ["orange",            8,[0xff,0x66,0x00,0xff],[0xff,0x66,0x00,0xff]], # 25
     ["unused",            0,[0x7f,0x7f,0x7f,0xff],[0x7f,0x7f,0x7f,0xff]], # 26
     ["unused",            0,[0x7f,0x7f,0x7f,0xff],[0x7f,0x7f,0x7f,0xff]], # 27
     ["unused",            0,[0x7f,0x7f,0x7f,0xff],[0x7f,0x7f,0x7f,0xff]], # 28
     ["unused",            0,[0x7f,0x7f,0x7f,0xff],[0x7f,0x7f,0x7f,0xff]], # 29
     ["unused",            0,[0x7f,0x7f,0x7f,0xff],[0x7f,0x7f,0x7f,0xff]], # 30
     ["unused",            0,[0x7f,0x7f,0x7f,0xff],[0x7f,0x7f,0x7f,0xff]], # 31
     ["unused",            0,[0x7f,0x7f,0x7f,0xff],[0x7f,0x7f,0x7f,0xff]], # 32
     ["trans_dk_blue",     9,[0x00,0x00,0x99,0xff],[0x7f,0x7f,0x7f,0x00]], # 33
     ["trans_green",      10,[0x00,0x50,0x18,0xff],[0x7f,0x7f,0x7f,0x00]], # 34
     ["unused",            0,[0x7f,0x7f,0x7f,0xff],[0x7f,0x7f,0x7f,0x00]], # 35
     ["trans_red",        12,[0xcc,0x00,0x00,0xff],[0x7f,0x7f,0x7f,0x00]], # 36
     ["unused",            0,[0x7f,0x7f,0x7f,0xff],[0x7f,0x7f,0x7f,0x00]], # 37  
     ["unused",            0,[0x7f,0x7f,0x7f,0xff],[0x7f,0x7f,0x7f,0x00]], # 38
     ["trans_grey",        8,[0x99,0x99,0x99,0xff],[0x7f,0x7f,0x7f,0x00]], # 39
     ["unused",            0,[0x7f,0x7f,0x7f,0xff],[0x7f,0x7f,0x7f,0x00]], # 40
     ["trans_lt_blue",     1,[0x99,0xc0,0xf0,0xff],[0x7f,0x7f,0x7f,0x00]], # 41
     ["trans_green_yellow",2,[0xcc,0xff,0x00,0xff],[0x7f,0x7f,0x7f,0x00]], # 42
     ["unused",            3,[0x7f,0x7f,0x7f,0xff],[0x7f,0x7f,0x7f,0x00]], # 43
     ["unused",            4,[0x7f,0x7f,0x7f,0xff],[0x7f,0x7f,0x7f,0x00]], # 44
     ["unused",            5,[0x7f,0x7f,0x7f,0xff],[0x7f,0x7f,0x7f,0x00]], # 45
     ["trans-yellow",      8,[0xf0,0xc4,0x00,0xff],[0x7f,0x7f,0x7f,0x00]], # 46
     ["clear",             8,[0xff,0xff,0xff,0xff],[0x7f,0x7f,0x7f,0x00]], # 47
     ["unused",            0,[0x7f,0x7f,0x7f,0xff],[0x7f,0x7f,0x7f,0x00]], # 48
     ["unused",            2,[0x7f,0x7f,0x7f,0xff],[0x7f,0x7f,0x7f,0x00]], # 49
     ["unused",            0,[0x7f,0x7f,0x7f,0xff],[0x7f,0x7f,0x7f,0x00]], # 50
     ["unused",            0,[0x7f,0x7f,0x7f,0xff],[0x7f,0x7f,0x7f,0x00]], # 51
     ["unused",            0,[0x7f,0x7f,0x7f,0xff],[0x7f,0x7f,0x7f,0x00]], # 52
     ["unused",            0,[0x7f,0x7f,0x7f,0xff],[0x7f,0x7f,0x7f,0x00]], # 53
     ["unused",            0,[0x7f,0x7f,0x7f,0xff],[0x7f,0x7f,0x7f,0x00]], # 54
     ["unused",            0,[0x7f,0x7f,0x7f,0xff],[0x7f,0x7f,0x7f,0x00]], # 55
     ["unused",            0,[0x7f,0x7f,0x7f,0xff],[0x7f,0x7f,0x7f,0x00]], # 56
     ["trans-orange",      8,[0xff,0x66,0x00,0xff],[0x7f,0x7f,0x7f,0x00]], # 57
     ["unused",            0,[0x7f,0x7f,0x7f,0xff],[0x7f,0x7f,0x7f,0x00]], # 58
     ["unused",            0,[0x7f,0x7f,0x7f,0xff],[0x7f,0x7f,0x7f,0x00]], # 59
     ["unused",            0,[0x7f,0x7f,0x7f,0xff],[0x7f,0x7f,0x7f,0x00]], # 60
     ["unused",            0,[0x7f,0x7f,0x7f,0xff],[0x7f,0x7f,0x7f,0x00]], # 61
     ["unused",            0,[0x7f,0x7f,0x7f,0xff],[0x7f,0x7f,0x7f,0x00]], # 62
     ["unused",            0,[0x7f,0x7f,0x7f,0xff],[0x7f,0x7f,0x7f,0x00]]  # 63
     );

my %colors_by_name = ();
my $i;
for ($i = int(@colors) - 1; $i >= 0; $i--) {
    my ($name, $reverse, $rgba, $dith) = @{$colors[$i]};
    $colors_by_name{$name} = [$i, $reverse, $rgba, $dith];
}

# Convert the LDraw color to an OpenGL color
sub convert_color {
    my ($ldcolor, $curcolor) = @_;
    
    if ($ldcolor !~ m|^\d+$|) {
	if (exists $colors_by_name{$ldcolor}) {
	    $ldcolor = $colors_by_name{$ldcolor}
	} else {
	    print STDERR "Bad color: $ldcolor\n";
	}
    }

    # 16 is the current color
    $ldcolor = $curcolor
	if $ldcolor == 16;

    # 24 is the complement of the current color
    $ldcolor = $colors[$curcolor]->[1]
	if $ldcolor == 24;

    my ($name, $reverse, $rgba, $dith) = @{$colors[$ldcolor]};
    my @color = @$rgba;

    my $c;
    foreach $c (@color) {
	$c /= 0xff;
    }

    return @color;
}

# Return true so we compile
1;
