#!/usr/local/bin/perl -w

# This provides a class for manipulating 1XN vectors

package Vector;

use strict;
use vars qw(@ISA @EXPORT @EXPORT_OK %EXPORT_TAGS $VERSION);

# Allow carping of errors
use Carp;

# Can overload with the overload package by uncommenting the following
#  and filling in the appropriate functions
# use overload ('+' => \&vector_add);

use Exporter;
$VERSION = 1.00;
@ISA = qw(Exporter);

@EXPORT      = qw();
@EXPORT_OK   = qw();
%EXPORT_TAGS = ();


=head1 NAME

Vector - Store and manipulate an 1XN vector (Subclass of Matrix)

=head1 SYNOPSIS

    N is the number of rows (or Y)
    
    my $vector =  Vector->new("-n" => 2,
			      -data => [1, 
			                3, 
					7, 
					2]
			      );
    print $vector->print;

    This should print:
        1
        3
        7
	2

=cut

=head2 Vector->new(%ARGS);

This accepts arguments in hash form:

   Vector->new("-n" => 2)

Which makes a new 1x2 vector with all elements zeroed.


The valid argument names are:

=over 4

=item C<-n>: N dimension (rows)

=item C<-data>: [ .. ] data information in an arrayref
if the data is larger than C<-n> then it is cut off, smaller data is 
zeroed

=back

The leading - is optional on args but recommended.  The C<-n>
are optional if data is specified in which case it will use the -data arg
to determine the size of the vector.  The vector will be created with 
the maximum possible size and zeroes fill in missing data.

Note that it is never okay to create a vector without specifying the size
in some way, either explicitly or giving data to infer the size from.

This returns a Vector object.

=cut

sub new {
    my $class = shift;
    
    # Get the class name if we were called through an object
    $class = ref $class if ref $class;

    # Set up memory for the object and bless it into the class
    my $self = {};
    bless($self, $class);

    # Initialize the object with the remaining arguments
    return $self->_init(@_);
}


=head2 $matrix->get($n);
=head2 $matrix->set($n, $value);

This will either get the value of matrix entry $m, $n.
Or it will set matrix entry $m, $n to $value and return the new value.

The first element in a matrix is 0, 0.  Note that N and M are never inside
the matrix, the last element is N-1, M-1.

=cut

sub get {
    my $self = shift;

    # If we were called through the class then we need to shift 
    # again to get the object
    $self = shift
	unless ref $self;

    my $n = shift;
    croak "insufficient arguments" 
	unless defined($n);

    # See if we were given the m argument and if we were get the right n
    $n = shift if @_;

    # Check the size
    croak "attempt to access an element outside the vector"
	unless $n < $self->N;

    # Call the function that does the work
    return $self->_get($n);
}

sub set {
    my $self = shift;

    # If we were called through the class then we need to shift 
    # again to get the object
    $self = shift
	unless ref $self;

    my $m = 0;
    my $n = shift;
    my $value = shift;
    croak "insufficient arguments" 
	unless defined($m) and defined($n) and defined($value);

    # See if we were given the m argument and if we were get the right n
    if (@_) {
	$n = $value;
	$value = shift;
    }

    # Check the size
    croak "attempt to access an element outside the vector"
	unless $n < $self->N;

    # Call the function that does the work
    return $self->_set($n, $value);
}

sub x {
    my $self = shift;
    return $self->get(0);
}
sub y {
    my $self = shift;
    return $self->get(1);
}
sub z {
    my $self = shift;
    return $self->get(2);
}


=head2 $matrix->N;

These return the dimensions of the matrix.  The dimensions can not 
change after creation.

=cut

sub N {
    my $self = shift;
    croak "instance method called on class" unless ref $self;
    return $self->{"N"};
}


=head2 $matrix->Nindices;

These return the all of the valid N indices so 0 .. N.

=cut

sub Nindices {
    my $self = shift;
    croak "instance method called on class" unless ref $self;
    return (0 .. ($self->{"N"}-1) );
}


=head2 $vector1->add($vector2)

This computes the result of $vector1 + $vector2.

Vectors to be added must be the same size.

The return value is a new vector with the sum of the two input vectors.

=cut

sub add {
    my $self = shift;
    my $other = shift;

    return $self->_iterate($other, sub { $_[0] + $_[1] } );
}


=head2 $vector1->subtract($vector2)

This computes the result of $vector1 - $vector2.

Vectors to be subtracted must be the same size.

The return value is a new vector with the result of the two input vectors.

=cut

sub subtract {
    my $self = shift;
    my $other = shift;

    return $self->_iterate($other, sub { $_[0] - $_[1] } );
}


=head2 $vector1->multiply($vector2)

This computes the result of scalar multiplication of the terms of 
$vector1 and  $vector2.  So a new vector where $vectorNew[N] = 
$vector1[N] * $vector2[N].

Vectors to be multiplied must be the same size.

The return value is a new vector with the result of the two input vectors.

=cut

sub multiply {
    my $self = shift;
    my $other = shift;

    return $self->_iterate($other, sub { $_[0] * $_[1] } );
}


=head2 $vector->scale(3);

This returns a new vector that is the result of multiplying each term
by the given scalar value.
    i.e. X => X * 3, Y = Y * 3 ...

=cut

sub scale {
    my $self = shift;
    my $value = shift;

    return $self->_iterate($self, sub { $_[0] * $value } );    
}


=head2 $vector->reverse;

This returns a new vector that is the reverse of the given vector.  
    i.e. X => -X, Y = -Y ...

=cut

sub reverse {
    my $self = shift;

    return $self->_iterate($self, sub { $_[0] * -1 } );    
}


=head2 $vector1->dot($vector2);

This will compute the dot product of $vector1 and $vector2.  The 
vectors must be the same size.

It will return a single value that is the computed dot product.

=cut

sub dot {
    my $self = shift;
    my $other = shift;
    croak "instance method called on class" unless ref $self;
    croak "no second vector supplied" unless ref $other;

    # Check the vector dimensions
    croak "Unable to compute the dot product of different sized vectors"
	unless $self->N == $other->N;
    
    # Loop through the vector and calculate the dot product
    my $result = 0;
    my $i = 0;
    my $sN = $self->N;
    for ($i = 0; $i < $sN; $i++) {
	$result += $self->_get($i) * $other->_get($i);
    }
    
    return $result;
}


=head2 $vector1->cross($vector2);

This will compute the cross product of $vector1 and $vector2.  The 
vectors must be the same size.

It will return a vector corresponding to the resulting cross product.

=cut

sub cross {
    my $s = shift;
    my $o = shift;
    croak "instance method called on class" unless ref $s;
    croak "no second vector supplied" unless ref $o;

    # Check the vector dimensions
    croak "Unable to compute the cross product of different sized vectors"
	unless $s->N == 3 and
	    $s->N == $o->N;
    
    # Set up the result vector
    my $result = Vector->new("-n" => $s->N);

    # And compute the result
    my $r;
    $r = $s->_get(1)*$o->_get(2) - $s->_get(2)*$o->_get(1);
    $result->_set(0, $r);

    $r = $s->_get(2)*$o->_get(0) - $s->_get(0)*$o->_get(2);
    $result->_set(1, $r);
    
    $r = $s->_get(0)*$o->_get(1) - $s->_get(1)*$o->_get(0);
    $result->_set(2, $r);
    
    return $result;
}


=head2 $vector->normalize();

This will normalize the vector to a length of 1

=cut

# Normalize the vector to a length of 1
sub normalize {
    my $vec = shift;

    my $len = 
	  $vec->get(0) * $vec->get(0)
	+ $vec->get(1) * $vec->get(1)
	+ $vec->get(2) * $vec->get(2);
    $len = sqrt($len);

    $vec->set(0, $vec->get(0) / $len);
    $vec->set(1, $vec->get(1) / $len);
    $vec->set(2, $vec->get(2) / $len);

    return $vec;
}


=head2 $vector1->data_asrow();

This will return a ref to the data as an array of 
scalars. i.e. (1, 3, 2) would be a possible return value.

B<BEWARE> that this may not be a copy of the data, this may be the raw
data itself so change it at your own risk.

=cut

sub data_asrow {
    my $self = shift;

    # Return the data if we can
    return @{$self->{"DATA"}} if $self->{"ONE_DIM"};

    # This should not happen but if it does copy the data
    my @a = ();
    my $i = 0;
    foreach $i ($self->Nindices) {
	$a[$i] = $self->_get($i);
    }

    return @a;
}


=head2 $vector->name

This returns the name of the object as a string.

=cut

sub name {
    my $self = shift;
    croak "instance method called on class" unless ref $self;
    
    return $self->{"NAME"};
}


=head2 $vector->print

This returns the contents of the vector with some attempt to pretty 
print it.  The format is not yet fixed.

=cut

sub print {
    my $self = shift;

    # If we were called through the class then we need to shift 
    # again to get the object
    $self = shift
	unless ref $self;
    croak "instance method called on class" unless ref $self;

    my $ret = "";
    
    $ret .= $self->name . " is " . 1 . " x " . $self->N . ": " .
	join(":", $self->data_asrow) . "\n";

    return $ret;
}


=head2 $vector->clone

This returns a clone of the vector with separate but identical data.

=cut

sub clone {
    my $self = shift;
    croak "instance method called on class" unless ref $self;

    return $self->new("-N" => $self->N, 
		      "-COPY" => $self->{"DATA"});
}


####
# PRIVATE functions, do not use unless you absolutely have to
#   I will not guarantee that the below will not change
####

# Iterate over a pair of matrices and apply the operation to both
sub _iterate {
    my $self = shift;
    my $other = shift;
    my $operation = shift;
    croak "instance method called on class" unless ref $self;
    croak "no second matrix supplied" unless ref $other;
    croak "no operation specified" unless ref $operation;

    # Check the matrix dimensions
    croak "Unable to apply operation to different sized vectors"
	unless $self->N == $other->N;

    # Create the new matrix
    my $result = $self->new("-N" => $self->N);

    # Do the operation
    my $n;
    my $sN = $self->N - 1;
    foreach $n (0 .. $sN) {
	my $value = &$operation($self->_get($n), $other->_get($n));
	$result->_set($n, $value);
    }

    return $result;
}

sub _get {
    return $_[0]->{'DATA'}[$_[1]];
}

sub _set {
    return ($_[0]->{'DATA'}[$_[1]] = $_[2]);
}

# Initialize the matrix
# Should probably be called by subclasses
sub _init {
    my $self = shift;
    
    # Set up the base arguments
    $self->{"N"} = undef;
    $self->{"DATA"} = undef;
    $self->{"NAME"} = undef;

    # Examine the arguments
    # Did we get any arguments?
    croak "No size was specified to the matrix constructor"
	if @_ == 0;

    # Must we copy the data?
    my $copy = 0;

    # Iterate through the arguments and populate the data
    my $arg;
    while (defined ($arg = shift) ) {
	$arg = uc $arg;
	$arg =~ s|\s||g;
	$arg =~ s|^-||;

	# Set the copy flag
	if ($arg eq "COPY") {
	    $copy = 1;
	    $arg = "DATA";
	}

	if (exists $self->{$arg}) {
	    carp "Defining $arg twice"
		if defined $self->{$arg};
	    $self->{$arg} = shift;
	} else {
	    carp "Bad argument $arg";
	}
    }

    # Check that we got what we needed and try to fill in anything that 
    # is missing
    # Check the name
    $self->{"NAME"} = $self->{"NAME"} || "Vector";
    
    # If N was not given then look at the data to determine the size
    $self->{"N"} = $self->{"N"} || @{$self->{"DATA"}};

    carp "N must be greater than zero"
	unless $self->N > 0;

    # Copy the data into our internal array if we have to
    if ($copy) {
	$self->{"DATA"} = [@{$self->{"DATA"}}];
    }
    
    # Define data if it was not specified
    unless (defined $self->{"DATA"}) {
	my $i;
	foreach $i ($self->Nindices) {
	    $self->{"DATA"}[$i] = 0;
	}
    }

    return $self;
}

1;
