#!/usr/local/bin/perl -w

# Enforce rigorous syntax rules
use strict;

# Set the path to the modules
use lib qw(/home/benb/perlld/Modules /home/fiji/rokenbok/perlld/Modules
	   ./Modules);

# Display errors from the viewpoint of the caller
use Carp;

# Allow access to the OpenGL 3D graphics routines
use OpenGL;

# Use my simple Vector class
use Vector;

# Provide an object oriented interface to the LDraw file contents
use LDraw;

# Include the polygon facing routines
#use CheckFacing;

# Include the displaylist routines
#use DisplayList;

# Include the display routines
use Render;

# Include the quaternion rotation functions
use Quat;

# Include the Tk modules
use Tk;
use Tk::LabEntry;

# Include our UI routines
use UI;


# These are the global variables
my $ld = undef;          # The LDraw object
my $obj_list = undef;    # The display list number
my $inited = 0;          # Our init state
my $rotation = 0;        # The current rotation of the object
my $curquat = [];        # The current quaternion
my $obj_offset = undef;  # The offset to recenter the object
my $obj_size = undef;    # The object size
my $max_dimension = 0;   # The diameter of the bounding sphere
my ($viewWidth, $viewHeight) = (300, 300); # The GL window size

# Our x, y, z position
my %pos = ('x' => 0, 'y' => 0, 'z' => 80);
my %rot = ('x' => 0, 'y' => 0, 'z' => 0);

# MAIN
{
    # Initialize the quaternion
    trackball($curquat, 0, 0, 0, 0);
    
    # Read the file into an object
    $ld = LDraw->new(shift);
    
    # Populate the object size & offsets
    my $max = $ld->get_bounding_points->{"MAX"};
    my $min = $ld->get_bounding_points->{"MIN"};
    $obj_size = $max->subtract($min);
    $obj_offset = $min->add($obj_size->scale(.5));

    # Work out the radius of the bounding sphere
    my $v;
    foreach $v ($max->data_asrow, $min->data_asrow) {
	$max_dimension = abs($v)
	    if abs($v) > $max_dimension;
    }


    # Calculate the display list
    #my $dl = DisplayList->new($ld);

    # Draw the object by walking it with drawing actions
    #correct_poly_facing($dl);
    #ray_poly_facing($dl);


    # Draw the UI
    my ($mw, $gl, $spin_b) = setup_ui($viewWidth, $viewHeight, \%pos, \%rot);
    $spin_b->configure(-command => [\&toggle_spin, $gl, $spin_b]);

    # Setup the custom bindings
    $gl->Tk::bind('GLT',"<Configure>",[\&init_callback, Ev('h'), Ev('w')]);
    $gl->Tk::bind('GLT',"<Visibility>",\&post_redraw);
    $gl->Tk::bind('GLT',"<Expose>",\&post_redraw);
    $gl->Tk::bind('GLT',"<ButtonPress-1>",[\&rot, 0, Ev('x'), Ev('y')]);
    $gl->Tk::bind('GLT',"<ButtonRelease-1>",[\&rot, 1, Ev('x'), Ev('y')]);
    $gl->Tk::bind('GLT',"<ButtonPress-3>",[\&pick_point, Ev('x'), Ev('y')]);
    $gl->Tk::bind('GLT',"<KeyPress>",[\&keyhit, Ev('K'), Ev('k'), Ev('T')]);
    $gl->Tk::bind('GLT',"<KeyRelease>",[\&keyhit, Ev('K'), Ev('k'), Ev('T')]);

    # Set the order that the bindings get called
    $gl->bindtags([ ref $gl, $gl, $gl->toplevel, 'all', 'GLT']);
    
    # And start the redraw process
    $gl->repeat(10, \&redraw);
    
  Tk::MainLoop();
}

# These functions share persistant private variables
BEGIN {
    my $shifted = 0;
    my $controlled = 0;

    my %key_bindings = 
	("Up" => sub { $pos{'z'} -= 5; post_repersp(); },
	 "Down" => sub { $pos{'z'} += 5; post_repersp(); },
	 "Left" => sub { $rot{'y'} -= 5; post_repersp(); },
	 "Right" => sub { $rot{'y'} += 5; post_repersp(); },

	 "Up-Shift" => sub { $rot{'x'} -= 5; post_repersp(); },
	 "Down-Shift" => sub { $rot{'x'} += 5; post_repersp(); },
	 "Left-Shift" => sub { $pos{'x'} -= 5; post_repersp(); },
	 "Right-Shift" => sub { $pos{'x'} += 5; post_repersp(); },
	 
	 "Up-Ctrl" => sub { $pos{'y'} -= 5; post_repersp(); },
	 "Down-Ctrl" => sub { $pos{'y'} += 5; post_repersp(); },
	 "Left-Ctrl" => sub { $rot{'z'} -= 5; post_repersp(); },
	 "Right-Ctrl" => sub { $rot{'z'} += 5; post_repersp(); },
	 );

    sub keyhit {
	my ($widget, $keysym, $keyval, $event) = @_;

	if ($event eq 'KeyPress') {
	    $shifted = 1 if $keysym =~ m|^Shift_[LR]$|;
	    $controlled = 1 if $keysym =~ m|^Control_[LR]$|;

	    # Process the key event
	    my $newsym = $keysym;
	    $newsym .= "-Shift" if $shifted;
	    $newsym .= "-Ctrl" if $controlled;
	    if (exists $key_bindings{$newsym}) {
		&{$key_bindings{$newsym}};
	    }
	}

	elsif ($event eq 'KeyRelease') {
	    $shifted = 0 if $keysym =~ m|^Shift_[LR]$|;
	    $controlled = 0 if $keysym =~ m|^Control_[LR]$|;
	}

    }
}

# These functions share persistant private variables
BEGIN {
    my $beginX;
    my $beginY;

    # $up of 0 is down, 1 is up and 2 is movement
    sub rot {
	my ($widget, $up, $x, $y) = @_;

	# When the user clicks in the window take the keyboard focus
	$widget->CanvasFocus;
    
	# Down
	if ($up == 0) {
	    stop_spinning();

	    $beginX = $x;
	    $beginY = $y;
	
	    $widget->Tk::bind('GLT',"<Motion>",[\&rot, 2, Ev('x'), Ev('y')]);
	    return;
	} 

	# Up
	if ($up == 1) {
	    # Unbind drawing on mouse movements
	    $widget->Tk::bind('GLT',"<Motion>", "");
	}

	# Up or movement
	my $lastquat = [];
	trackball($lastquat,
		  (2.0*$beginX - $viewWidth) / $viewWidth,
		  -($viewHeight - 2.0*$beginY) / $viewHeight,
		  (2.0*$x - $viewWidth) / $viewWidth,
		  -($viewHeight - 2.0*$y) / $viewHeight
		  );
	add_quats($lastquat, $curquat, $curquat);

	post_redraw($widget);
    }
}

# These functions share persistant private variables
BEGIN {
    my $spinning = 0;
    my $spin_id = undef;
    my $spin_widget = undef;
    my $button_widget = undef;

    sub start_spinning {
	if (not defined $spin_widget) {
	    stop_spinning();
	    return;
	}
	$spinning = 1;
	$button_widget->configure(-text => "Stop Spin");
	$spin_id = $spin_widget->repeat(300, [\&do_spin]);
    }

    sub do_spin {
	$rotation = ($rotation+5) % 360;
	post_redraw($spin_widget);
    }

    sub stop_spinning {
	$spinning = 0;
	$button_widget->configure(-text => "Start Spin")
	    if defined $button_widget;
	$spin_id->cancel 
	    if defined $spin_id;
	$spin_id = undef;
    }

    sub toggle_spin {
	my $gl = shift;
	my $button = shift;

	$spin_widget = $gl;
	$button_widget = $button;

	if ($spinning) {
	    stop_spinning;
	} else {
	    start_spinning;
	}
    }
}

# These functions share persistant private variables
BEGIN {
    my $need_redraw = 1;     # Do we need to redraw the object?
    my $need_repersp = 1;    # Do we need to recalc the perspective?

    sub post_redraw {
	$need_redraw = 1;
    }

    sub post_repersp {
	$need_repersp = 1;
	post_redraw();
    }

    sub redraw {
	# Do we need to redraw?
	return unless $inited and $need_redraw;
	$need_redraw = 0;
	
	# Draw the object
	draw_body();
	
	# Show our drawing
	glXSwapBuffers;
    }

    sub draw_body {
	# Glear the window
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
	
	# See if we need to recalculate the perspective
	my $change_perspective = $need_repersp;
	$need_repersp = 0;
	
	# Set the projection and save the current values
	if ($change_perspective) {
	    glMatrixMode(GL_PROJECTION);
	    glPopMatrix;
	    glPushMatrix();
	    set_perspective();
	    glMatrixMode(GL_MODELVIEW);
	}
	
	# Save the current modelview values
	glPushMatrix();
	
	# Set the eye location 
	glRotatef(180, 0, 1, 0);         # Turn around
	glRotatef($rot{'x'}, 1, 0, 0);   # Around X: Up/down
	glRotatef($rot{'y'}, 0, 1, 0);   # Around Y: Left/right
	glRotatef($rot{'z'}, 0, 0, 1);   # Around Z: Pitch
	
	# Move to the eye point 
	glTranslatef($pos{'x'}, $pos{'y'}, $pos{'z'});
	
	# Center the object
	#glTranslatef(-$obj_offset->x,-$obj_offset->y,-$obj_offset->z);
	
	# Apply the rotation matrix based on the quaternion
	my $m = [];
	build_rotmatrix($m, $curquat);
	glMultMatrixf(pack("f16",@{$m->[0]}, @{$m->[1]}, 
			   @{$m->[2]}, @{$m->[3]}));
	
	# Rotate matrix around a YZ vector by $angle
	glRotatef($rotation, 0.0, 1.0, 1.0); 
	
	glCallList($obj_list);       # Draw the object
	
	glPopMatrix();               # Restore matrix values
	
    	check_for_glerror();
    }
}

sub pick_point {
    my ($widget, $x, $y) = @_;

    # Make a buffer of the correct size
    my $bufsize = 1024;
    my @buf; my $i;
    for ($i = 0; $i < $bufsize; $i++) {
	$buf[$i] = 0;
    }
    my $selectbuf = pack("I$bufsize", @buf);

    # Use the selection buffer and enter select mode
    glSelectBuffer($bufsize, $selectbuf);
    glRenderMode(GL_SELECT);

    # Set up the name stack
    glInitNames();
    glPushName(0);
    glPushName(0);

    # Get the current viewport information
    my $viewport = pack('i4', 0, 0, 0, 0);
    glGetIntegerv(GL_VIEWPORT, $viewport);

    # Set the projection up
    glMatrixMode(GL_PROJECTION); {
	glPushMatrix();     # Save the previous state
	glLoadIdentity();   # Clear the new matrix
	# Set our picking area
	gluPickMatrix($x, $viewHeight - $y, 1, 1, $viewport);
	post_repersp;       # Post a recalculation of the perspective
	glPushMatrix();     # draw_body will do a pop first thing
    } glMatrixMode(GL_MODELVIEW);
    
    # Draw the object
    draw_body();

    # Restore the projection
    glMatrixMode(GL_PROJECTION); {
	glPopMatrix();      # Remove the draw_body push
	glPopMatrix();      # Remove our changes too
    }glMatrixMode(GL_MODELVIEW);

    glFlush();              # Force the changes to be done before continuing

    # Get the hits when we switch out of select mode
    my $hits = glRenderMode(GL_RENDER);

    # Process the hits and color the selected object
    if ($hits > 0) {
	my ($file, $line) = 
	    processhits($hits, unpack("I$bufsize", $selectbuf));
	set_color($ld, $file, $line, 4);
	post_redraw;
    }   
}

sub init_callback {
    my ($widget, $h, $w) = @_;
	

    # Make sure we do not init the following multiple times
    #return if $inited;
    $inited = 1;

    # Calculate the window height & width
    $viewHeight = $h if defined $h;
    $viewWidth = $w if defined $w;

    # Execute the GL stuff to set up the window
    gl_init($widget, $viewHeight, $viewWidth);

    # Make a display list
    rewalk_list();
}

sub set_perspective {
    # Calculate the near & far clipping planes
    my $len = $pos{'x'} * $pos{'x'} +
	$pos{'y'} * $pos{'y'} +
	$pos{'z'} * $pos{'z'};
    
    $len = sqrt($len);

    my $near = $len - $max_dimension;
    my $far = $len + $max_dimension;

    $near = 1 if $near <= 0;
    $far = 1 if $far <= 0;

    # Set the projection
    gluPerspective(40,                        # Field of View in degrees
		   $viewWidth / $viewHeight,  # Aspect ratio
		   $near,                     # Z near
		   $far,                      # Z far
		   );

}

sub rewalk_list {
    $obj_list = walk_all_display_lists($ld);

    post_repersp();
}
